using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Diagnostics;
using Editroid.Graphic;
using System.Collections;
using Editroid.ROM;
using System.Drawing.Imaging;

namespace Editroid.Graphic
{
	/// <summary>
	/// Renders a Metroid screen.
	/// </summary>
	public class ScreenDisplay
	{
		byte[,] tiles = new byte[32, 30];
		byte[,] colors = new byte[32, 30];
        Physics[,] physics = new Physics[32, 30];

        Level _Level;
        

		/// <summary>
		/// Gets/sets which level to obtain screen data from.
		/// </summary>
		public Level Level {
			get { return _Level; }
			set { _Level = value; }
		}

        private GameItem item;

        /// <summary>
        /// Get/set an item to be drawn when the screen is rendered.
        /// </summary>
        public GameItem GameItem {
            get { return item; }
            set { item = value; }
        }


        private byte defaultPal;

        /// <summary>
        /// Get/set the default palette.
        /// </summary>
        public byte DefaultPalette {
            get { return defaultPal; }
            set { defaultPal = value; }
        }


		/// <summary>
		/// Clears the screen and sprite data.
		/// </summary>
		public void Clear() {
			for(int i = 0; i < 960; i++) {
				tiles[i % 32, i / 32] = 0xFF;
                colors[i % 32, i / 32] = defaultPal;// Palette.Palette0;
                physics[i % 32, i / 32] = Physics.Air;
			}

            sprites.Clear();
            spriteLocations.Clear();
		}


		/// <summary>
		/// Draws an object to this screen display.
		/// </summary>
		/// <param name="obj">An instance of a combo. Palette and position information
		/// is specified in the ScreenObject.</param>
		public void DrawObject(ScreenObject obj) {
			int TileX = obj.X * 2;
			int ScreenY = obj.Y * 2;

			//Struct objDefinition = Level.GetStruct(obj.Index);
			byte[][] ComboData = Level.GetStruct(obj.Index).Combos; //objDefinition.Tiles;
			for(int ComboY = 0; ComboY < ComboData.Length; ComboY++) {
				byte[] ComboRow = ComboData[ComboY];
				int ScreenX = TileX;
				for(int ComboX = 0; ComboX < ComboRow.Length; ComboX++) {
					Combo c = Level.GetCombo(ComboRow[ComboX] & (byte)63);
					if(ComboRow[ComboX] != Struct.EmptyTile &&
                        ScreenX <= 31 && ScreenY <= 29) {



						// Apply tiles
                        byte combo;
                        // Top-left
                        combo = c.GetByte(0);
						tiles[ScreenX, ScreenY] = combo;
                        physics[ScreenX, ScreenY] = GetPhysics(combo);
                        // Top-Right
                        combo = c.GetByte(1);
                        tiles[ScreenX + 1, ScreenY] = combo;
                        physics[ScreenX + 1, ScreenY] = GetPhysics(combo);
                        // Bottom-Left
                        combo = c.GetByte(2);
                        tiles[ScreenX, ScreenY + 1] = combo;
                        physics[ScreenX, ScreenY + 1] = GetPhysics(combo);
                        // Bottom-Right
                        combo = c.GetByte(3);
                        tiles[ScreenX + 1, ScreenY + 1] = combo;
                        physics[ScreenX + 1, ScreenY + 1] = GetPhysics(combo);

						// Apply colors
                        if(obj.PalData != defaultPal)
						    colors[ScreenX, ScreenY] =
						    colors[ScreenX + 1, ScreenY] =
						    colors[ScreenX, ScreenY + 1] =
						    colors[ScreenX + 1, ScreenY + 1] =
    							obj.PalData;
					}
					ScreenX += 2;
				} // col

				ScreenY += 2;
			} // row
        }

        /// <summary>
        /// Gets the physics of a tile.
        /// </summary>
        /// <param name="tile">The index of a tile.</param>
        /// <returns>A Physics value for the physics of a tile.</returns>
        public static Physics GetPhysics(byte tile) {
            if(tile < 0x70)
                return Physics.Solid;
            if(tile > 0xA1)
                return Physics.Air;
            if(tile == 0xA0)
                return Physics.Door;
            if(tile == 0xA1)
                return Physics.DoorHorizontal;
            if(tile < 0x98) 
                return Physics.Breakable;

            return Physics.Solid;
        }

        private List<Sprite> sprites = new List<Sprite>();
        private List<Point> spriteLocations = new List<Point>();
        private List<byte> spritePalettes = new List<byte>();

        /// <summary>
        /// Adds a sprite to a batch to be rendered to the output image.
        /// </summary>
        /// <param name="sprite">The graphic sprite.</param>
        /// <param name="X">The location of the sprite.</param>
        /// <param name="Y">The location of the sprite.</param>
        /// <param name="palette">The index of the palette to render with.</param>
        public void AddSprite(Sprite sprite, int X, int Y, byte palette) {
            sprites.Add(sprite);
            spriteLocations.Add(new Point(X, Y));
            spritePalettes.Add(palette);
        }



        /// <summary>
        /// Draws an object to this screen display using the fourth palette table
        /// after the specified palette table to indicate that this object is selected
        /// </summary>
        /// <param name="obj">An instance of a combo. Palette and position information
        /// is specified in the ScreenObject.</param>
        public void DrawObjectSelected(ScreenObject obj) {
            int TileX = obj.X * 2;
            int ScreenY = obj.Y * 2;

            //Struct objDefinition = Level.GetStruct(obj.Index);
            byte[][] ComboData = Level.GetStruct(obj.Index).Combos; //objDefinition.Tiles;
            for(int ComboY = 0; ComboY < ComboData.Length; ComboY++) {
                byte[] ComboRow = ComboData[ComboY];
                int ScreenX = TileX;
                for(int ComboX = 0; ComboX < ComboRow.Length; ComboX++) {
                    Combo c = Level.GetCombo(ComboRow[ComboX]);

                    if(ComboRow[ComboX] != Struct.EmptyTile &&
                        ScreenX <= 31 && ScreenY <= 29) {
                        // Apply tiles
                        byte combo;
                        // Top-left
                        combo = c.GetByte(0);
                        tiles[ScreenX, ScreenY] = combo;
                        physics[ScreenX, ScreenY] = GetPhysics(combo);
                        // Top-Right
                        combo = c.GetByte(1);
                        tiles[ScreenX + 1, ScreenY] = combo;
                        physics[ScreenX + 1, ScreenY] = GetPhysics(combo);
                        // Bottom-Left
                        combo = c.GetByte(2);
                        tiles[ScreenX, ScreenY + 1] = combo;
                        physics[ScreenX, ScreenY + 1] = GetPhysics(combo);
                        // Bottom-Right
                        combo = c.GetByte(3);
                        tiles[ScreenX + 1, ScreenY + 1] = combo;
                        physics[ScreenX + 1, ScreenY + 1] = GetPhysics(combo);

                        // Apply colors
                        colors[ScreenX, ScreenY] =
                        colors[ScreenX + 1, ScreenY] =
                        colors[ScreenX, ScreenY + 1] =
                        colors[ScreenX + 1, ScreenY + 1] =
                            (byte)(obj.PalData + 4);
                    }
                    ScreenX += 2;
                } // col

                ScreenY += 2;
            } // row
        } // DrawObject

		/// <summary>
		/// Constructs the screen data in memory.
		/// </summary>
		/// <param name="index">The index of the screen to draw.</param>
		/// <param name="SelectedObject">The object that is currently selected.</param>
		/// <remarks>This function does not render the screen. It prepares the 
		/// data necessary to render the screen.</remarks>
		public void DrawScreen(int index, int SelectedObject) {
			Screen screen = Level.GetScreen(index);
			ScreenObject[] objects = screen.Objects;

			for(int i = 0; i < objects.Length; i++) {
				if(i == SelectedObject)
					DrawObjectSelected(objects[i]);
				else
					DrawObject(objects[i]);
			}
		}

		/// <summary>
		/// Constructs the screen data in memory.
		/// </summary>
		/// <param name="index">The index of the screen to draw.</param>
		/// <param name="SelectedObject">The object that is currently selected.</param>
		/// <remarks>This function does not render the screen. It prepares the 
		/// data necessary to render the screen.</remarks>
		public void DrawScreen(int index, ScreenObject SelectedObject) {
            if(Level == null) return;

            Screen screen = Level.GetScreen(index);
			ScreenObject[] objects = screen.Objects;

			for(int i = 0; i < objects.Length; i++) {
				if(objects[i].Equals(SelectedObject))
					DrawObjectSelected(objects[i]);
				else
					DrawObject(objects[i]);
			}
		}


		/// <summary>
		/// Constructs the screen data in memory.
		/// </summary>
		/// <param name="index">The index of the screen to draw.</param>
		/// <remarks>This function does not render the screen. It prepares the 
		/// data necessary to render the screen.</remarks>
		public void DrawScreen(int index) {
			DrawScreen(index, -1);
		}

		/// <summary>
		/// Renders a screen using GDI+.
		/// </summary>
		/// <param name="g">GDI+ graphic destination.</param>
		/// <remarks>This overload performs very poorly due to GDI+'s very poor
		/// optimizations for paletted bitmaps. This overload also does not render
		/// sprites in the screen, mainly because of the palette limitations of GDI+.</remarks>
		public void Render(Graphics g) {
			Rectangle source = new Rectangle(0, 0, 8, 8);
			Rectangle dest = source;

			PatternTable patTable = Level.Patterns;
			Bitmap patterns = patTable.Patterns;

			//patTable.Palette = palettes[colors[0, 0]];

			for(int x = 0; x < 32; x++) {
				dest.Y = 0;
				for(int y = 0; y < 32; y++) {
					//patTable.Palette = palettes[colors[x, y]];
					source.X = tiles[x, y] * 8;
					g.DrawImage(patterns, dest, source, GraphicsUnit.Pixel);
					dest.Y += 8;
					//if(x == 0 && y == 8) System.Windows.Forms.MessageBox.Show(tiles[x, y].ToString());
				}
				dest.X += 8;
			}
		}

		/// <summary>
		/// Renders the screen in memory with no sprites or doors.
		/// </summary>
		/// <param name="b">The Blitter object used for rendering.</param>
		/// <param name="dest">The bitmap to render to.</param>
		public void Render(Blitter b, Bitmap dest) {
			Render(b, dest, null, null, false);	
		}

        /// <summary>
        /// Applies the current level's (specified by the Level property) palette
        /// to a Bitmap, generally the Bitmap that is being rendered to.
        /// </summary>
        /// <param name="dest">The indexed bitmap to apply a color table to.</param>
        /// <param name="alt">If true, the level's alternate background palette will be used instead of the normal palette.</param>
        public void ApplyPalette(Bitmap dest, bool alt) {
            NesPalette pal = alt ?
                _Level.BgAltPalette :
                _Level.BgPalette;

            ColorPalette destPal = dest.Palette;
            pal.ApplyTable(destPal.Entries);
            dest.Palette = destPal;
        }

		/// <summary>
		/// Renders the screen in memory, additionally rendering any doors and sprites specified.
		/// </summary>
		/// <param name="b">The Blitter object used for rendering.</param>
		/// <param name="dest">The bitmap to render to.</param>
		/// <param name="sprites">An array of sprites to be rendered, or null if there are no sprites.</param>
        /// <param name="doors">An array of doors to render, or null if there are no doors.</param>
        /// <param name="showPhysics">If true then the screen will be drawn such that the phyics of the tiles are appearent. Sprites, doors, items, and the selection will be omitted.</param>
        public void Render(Blitter b, Bitmap dest, ScreenEnemy[] sprites, ScreenDoor[] doors, bool showPhysics) {
            //PatternTable patTable = Level.Patterns;
            //Bitmap patterns = patTable.Patterns;
            b.Begin(Level.Patterns.Patterns, dest);

            //patTable.Palette = palettes[colors[0, 0]];
            if(showPhysics) {
                RenderPhysics(b);
            } else {
                for(int x = 0; x < 32; x++) {
                    for(int y = 0; y < 30; y++) {
                        int tile = tiles[x, y];
                        b.BlitTile(tile, x, y, colors[x, y]);
                    }
                }

                b.ChangeSource(Level.SpritePatterns.Patterns);

                if(sprites != null) {
                    for(int i = 0; i < sprites.Length; i++) {
                        ScreenEnemy s = sprites[i];
                        Sprite enemySprite = Level.GetSprite(s.EnemyType);
                        byte pal = 0xA;
                        //if(selectedEnemy == i) pal = 0;
                        if(s.SpritePal == 8) pal = 0x8;

                        enemySprite.Draw(b, s.X * 2, s.Y * 2, pal);
                        if(s.Respawn)
                            ScreenEnemy.RespawnSprite.Draw(b, s.X * 2, s.Y * 2, 0);
                        if(selectedEnemy == i) {
                            Rectangle spriteBounds = enemySprite.Measure();
                            spriteBounds.X += s.X * 16;
                            spriteBounds.Y += s.Y * 16;
                            b.DrawRect(spriteBounds, NesPalette.HighlightEntry);
                        }
                    }
                }
                if(doors != null) {
                    foreach(ScreenDoor d in doors) {
                        byte pal = 9;
                        if(d.Type == DoorType.Missile) pal = 8;
                        else if(d.Type == DoorType.TenMissile) pal = 10;

                        if(d.Side == DoorSide.Left) {
                            Graphic.Sprite.LeftDoor.Draw(b, 2, 0xA, pal);
                        } else {
                            Graphic.Sprite.RightDoor.Draw(b, 0x1D, 0xA, pal);
                        }
                    }
                }
            }

            if(!showPhysics) {
                if(item != null)
                    item.Draw(b);


                for(int i = 0; i < this.sprites.Count; i++) {
                    this.sprites[i].Draw(b, spriteLocations[i].X, spriteLocations[i].Y, spritePalettes[i]);
                }

            }

            b.End();
        }

        private void RenderPhysics(Blitter b) {
            for(int x = 0; x < 32; x++) {
                for(int y = 0; y < 30; y++) {
                    int tile = tiles[x, y];

                    // Don't show non-solid tiles
                    if(physics[x, y] == Physics.Air)
                        tile = 0xFF;

                    b.BlitTile(tile, x, y, (byte)(colors[x, y] % 4));

                    if(physics[x, y] == Physics.Breakable) {
                        b.DrawDitherTile(x, y, 0);
                    }
                }
            }
        }
        private int selectedEnemy;

        /// <summary>
        /// Gets or sets the index of the selected enemy.
        /// </summary>
        public int SelectedEnemy {
            get { return selectedEnemy; }
            set { selectedEnemy = value; }
        }

	}


}
