using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Editroid.Graphic
{
	/// <summary>
	/// Contains a definition of an image composed of graphic tiles
	/// </summary>
	public class Sprite
	{
		byte[] data;
		private int _Width = 2;
		/// <summary>
		/// Gets/sets the width of this sprite in tiles (eight pixel units)
		/// </summary>
        public int Width { get { return 2; } set { _Width = value; } }
        //public int Width { get { return _Width; } set { _Width = value; } }
		private int _Height = 2;
		/// <summary>
		/// Gets/sets the height of this sprite in tiles (eight pixel units)
		/// </summary>
		public int Height { get { return _Height; } set { _Height = value; } }
	

		/// <summary>
		/// Creates a sprite from the specified sprite data
		/// </summary>
		/// <param name="spriteData">Data used for sprite</param>
		public Sprite(byte[] spriteData) {
			data = spriteData;
		}
		/// <summary>
		/// Creates a sprite from the specified sprite data
		/// </summary>
		/// <param name="spriteData">Data used for sprite</param>
		/// <param name="width">Width of sprite in tiles (eight pixel units)</param>
		/// <param name="height">Height of sprite in tiles (eight pixel units)</param>
		public Sprite(byte[] spriteData, int width, int height) {
			data = spriteData;
			_Width = width;
			_Height = height;
		}

		/// <summary>
		/// Creates an array of sprites from an array of sprite data arrays.
		/// </summary>
        /// <param name="data">Array of sprite data arrays.</param>
        /// <param name="sizes">Sprite sizes</param>
		/// <returns>An array of sprites.</returns>
		public static Sprite[] FromArray(byte[][] data, Size[] sizes){
			Sprite[] sprites = new Sprite[data.Length];
			bool DefaultSize = sizes == null;

			for(int i = 0; i < data.Length; i++) {
				if(DefaultSize)
					sprites[i] = new Sprite(data[i], 2, 2);
				else
					sprites[i] = new Sprite(data[i], sizes[i].Width, sizes[i].Height);
			}

			return sprites;
		}

		/// <summary>
		/// Draws a sprite using a blitter.
		/// </summary>
		/// <param name="b">The blitter to use to render the sprite.</param>
		/// <param name="X">The X-coordinate of the sprite.</param>
		/// <param name="Y">The Y-coordinate of the sprite.</param>
		/// <param name="pal">The number of the palette to render the sprite with.</param>
		/// <remarks>The blitter must be open for rendering before this function
		/// can be called.</remarks>
		public void Draw_Unused(Blitter b, int X, int Y, byte pal) {
			int row = 0;
			int tileX = X;
			int tileY = Y;
			Blitter.FlipFlags flip = 0;
			byte palette = pal;
			int Offset = 0;

			for(int i = 0; i < data.Length; i++) {
				palette = pal;
				if(data[i] >= 0xB0 && data[i] <= 0xBF) { // If data is macro, intepret
					switch(data[i]) {
						case SpriteMacros.FlipX:
							flip |= Blitter.FlipFlags.Horizontal;
							break;
						case SpriteMacros.FlipY:
							flip |= Blitter.FlipFlags.Vertical;
							break;
						case SpriteMacros.MoveD4:
							Offset += 1024;
							break;
						case SpriteMacros.MoveU4:
							Offset -= 1024;
							break;
						case SpriteMacros.MoveL4:
							Offset -= 4;
							break;
						case SpriteMacros.MoveR4:
							Offset += 4;
							break;
						case SpriteMacros.NextPal:
							palette = (byte)((palette + 1) % 4);
							break;
						case SpriteMacros.NextRow:
							row++;
							Offset = 0;
							tileX = X;
							tileY = Y + row;
							break;
					}
				} else { // if not a macro, it is a tile. Render.
					if(Offset + tileX + tileY * 0x800 > 0
						&& Offset + tileX + tileY * 0x800 < 0xF988) {

						b.BlitTileTransparent(data[i], tileX, tileY, palette, flip); //, Offset);
						tileX++;
					}
					flip = 0;
					palette = pal;
				}
			}
		}

        /// <summary>
        /// Draws a sprite using a blitter.
        /// </summary>
        /// <param name="b">The blitter to use to render the sprite.</param>
        /// <param name="X">The X-coordinate of the sprite.</param>
        /// <param name="Y">The Y-coordinate of the sprite.</param>
        /// <param name="pal">The number of the palette to render the sprite with.</param>
        /// <remarks>The blitter must be open for rendering before this function
        /// can be called.</remarks>
        public void Draw(Blitter b, int X, int Y, byte pal) {
            int row = 0;
            int tileX = X;
            int tileY = Y;
            Blitter.FlipFlags flip = 0;
            byte palette = pal;
            int Offset = 0;
            int offsetX = 0;
            int offsetY = 0;

            for(int i = 0; i < data.Length; i++) {
                palette = pal;
                if(data[i] >= 0xB0 && data[i] <= 0xBF) { // If data is macro, intepret
                    switch(data[i]) {
                        case SpriteMacros.FlipX:
                            flip |= Blitter.FlipFlags.Horizontal;
                            break;
                        case SpriteMacros.FlipY:
                            flip |= Blitter.FlipFlags.Vertical;
                            break;
                        case SpriteMacros.MoveD4:
                            Offset += 1024;
                            offsetY += 4;
                            break;
                        case SpriteMacros.MoveU4:
                            Offset -= 1024;
                            offsetY -= 4;
                            break;
                        case SpriteMacros.MoveL4:
                            Offset -= 4;
                            offsetX -= 4;
                            break;
                        case SpriteMacros.MoveR4:
                            Offset += 4;
                            offsetX += 4;
                            break;
                        case SpriteMacros.NextPal:
                            palette = (byte)((palette + 1) % 4);
                            break;
                        case SpriteMacros.NextRow:
                            row++;
                            Offset = 0;
                            tileX = X;
                            tileY = Y + row;
                            offsetX = offsetY = 0;
                            break;
                    }
                } else { // if not a macro, it is a tile. Render.
                    if((tileX * 8 + offsetX) >= 0 &&
                        (tileX * 8 + offsetX + 7) < 256 &&
                        (tileY * 8 + offsetY) >= 0 &&
                        (tileY * 8 + offsetY + 7) < 240) {

                        b.BlitTileTransparent_Pixel(data[i], tileX * 8 + offsetX, tileY * 8 + offsetY, palette, flip); //, Offset);
                        tileX++;
                    }
                    flip = 0;
                    palette = pal;
                }
            }
        }

        /// <summary>
        /// Measures a sprite.
        /// </summary>
        /// <returns>A rectangle that represents the bounds of a sprite
        /// relative to its zero coordinate.</returns>
        public Rectangle Measure() {
            int row = 0;
            int tileX = 0;
            int tileY = 0;
            Rectangle result = new Rectangle();

            for(int i = 0; i < data.Length; i++) {
                if(data[i] >= 0xB0 && data[i] <= 0xBF) { // If data is macro, intepret
                    switch(data[i]) {
                        case SpriteMacros.FlipX:
                            break;
                        case SpriteMacros.FlipY:
                            break;
                        case SpriteMacros.MoveD4:
                            tileY += 4;
                            break;
                        case SpriteMacros.MoveU4:
                            tileY -= 4;
                            break;
                        case SpriteMacros.MoveL4:
                            tileX -= 4;
                            break;
                        case SpriteMacros.MoveR4:
                            tileX += 4;
                            break;
                        case SpriteMacros.NextPal:
                            break;
                        case SpriteMacros.NextRow:
                            row++;
                            tileX = 0;
                            tileY = row * 8;
                            break;
                    }
                } else { // if not a macro, it is a tile. Render.
                    // Expand bounding rectangle to include rendered tile
                    if(tileX < result.Left) {
                        result.Width += result.Left - tileX;
                        result.X = tileX;
                    }
                    if(tileX + 8 > result.Right) {
                        result.Width += tileX + 8 - result.Right;
                    }
                    if(tileY < result.Top) {
                        result.Height += result.Top- tileY;
                        result.Y = tileY;
                    }
                    if(tileY + 8 > result.Bottom) {
                        result.Height += tileY + 8 - result.Bottom;
                    }

                    tileX += 8;
                }
            }

            return result;
        }

		#region Sprite Defs
		/// <summary>
		/// The sprite used for an opening door on the left edge of a screen.
		/// </summary>
		static public Sprite LeftDoor = new Sprite(new byte[] {
			0x0F, SpriteMacros.NextRow,
			0x1F, SpriteMacros.NextRow,
			0x2F, SpriteMacros.NextRow,
			SpriteMacros.FlipY, 0x2F, SpriteMacros.NextRow,
			SpriteMacros.FlipY, 0x1F, SpriteMacros.NextRow,
			SpriteMacros.FlipY, 0x0F
		}, 1, 6);

		/// <summary>
		/// The sprite used for an opening door on the right edge of a screen.
		/// </summary>
		static public Sprite RightDoor = new Sprite(new byte[] {
			SpriteMacros.FlipX, 0x0F, SpriteMacros.NextRow,
			SpriteMacros.FlipX, 0x1F, SpriteMacros.NextRow,
			SpriteMacros.FlipX, 0x2F, SpriteMacros.NextRow,
			SpriteMacros.FlipX, SpriteMacros.FlipY, 0x2F, SpriteMacros.NextRow,
			SpriteMacros.FlipX, SpriteMacros.FlipY, 0x1F, SpriteMacros.NextRow,
			SpriteMacros.FlipX, SpriteMacros.FlipY, 0x0F
		}, 1, 6);

        /// <summary>
        /// Returns true if this tile has no data.
        /// </summary>
        public bool IsBlank { get { return this.data.Length == 0; } }
		/// <summary>
		/// Enemy sprites used in Brinstar.
		/// </summary>
		static public Sprite[] BrinstarSprites = Sprite.FromArray(new byte[][]{
			new byte[]{},
			new byte[]{},
			new byte[]{0xD0, 0xD1},
			new byte[]{0xC0, 0xC1},
			new byte[]{SpriteMacros.MoveR4,SpriteMacros.MoveR4, 0xCE, SpriteMacros.NextRow,
						SpriteMacros.MoveR4,0xDE, 0xDF, SpriteMacros.NextRow,
						SpriteMacros.MoveR4,0xEE, SpriteMacros.FlipX, 0xEE},
			new byte[]{0xCC, 0xCD, SpriteMacros.NextRow, 0xDC, 0xDD},
			new byte[]{SpriteMacros.MoveL4, SpriteMacros.MoveU4, 0xE4, SpriteMacros.MoveD4, 0xE2, SpriteMacros.MoveU4, SpriteMacros.FlipX, 0xE4, SpriteMacros.NextRow,
						0xE3, SpriteMacros.FlipX, 0xE3},
			new byte[]{SpriteMacros.MoveL4, SpriteMacros.MoveL4, 0xC2, 0xC3, SpriteMacros.NextRow, SpriteMacros.MoveL4, SpriteMacros.MoveL4, 0xD2, 0xD3}
		}, new Size[]{
			new Size(0,0),
			new Size(0,0),
			new Size(2,1),
			new Size(2,1),
			new Size(2,3),
			new Size(2,2),
			new Size(2,2),
			new Size(2,2),
			});
		/// <summary>
		/// Enemy sprites used in Norfair.
		/// </summary>
		static public Sprite[] NorfairSprites = Sprite.FromArray(new byte[][]{
			new byte[]{SpriteMacros.MoveL4, 0xCE, 0xCF, SpriteMacros.FlipX, 0xCE, SpriteMacros.NextRow,
						0xDE, SpriteMacros.FlipX, 0xDE},
			new byte[]{},
			new byte[]{0xF6, 0xF7},
			new byte[]{},
			new byte[]{},
			new byte[]{},
			new byte[]{0xCC, 0xCD, SpriteMacros.NextRow, 0xDC, 0xDD},
			new byte[]{0xC2, 0xC3, SpriteMacros.NextRow, 0xD2, 0xD3},
			new byte[]{},
			new byte[]{},
			new byte[]{},
			new byte[]{0xEA, SpriteMacros.FlipX, 0xEA, SpriteMacros.NextRow,
						0xFA, SpriteMacros.FlipX, 0xFA},
			new byte[]{0xEE, 0xEF, SpriteMacros.NextRow,
						SpriteMacros.FlipX, SpriteMacros.FlipY, 0xEF, SpriteMacros.FlipY, 0xEF},
			new byte[]{0xC8, 0xC9, SpriteMacros.NextRow,
						0xD8, 0xD9, SpriteMacros.NextRow,
						0xE8, 0xE9, SpriteMacros.NextRow,
						0xF8, 0xF9, SpriteMacros.NextRow},
			new byte[]{SpriteMacros.MoveR4, SpriteMacros.MoveR4, SpriteMacros.MoveR4,  0XC0}
		}, new Size[]{
			new Size(2, 2),
			new Size(0, 0),
			new Size(2, 1),
			new Size(0, 0),
			new Size(0, 0),
			new Size(0, 0),
			new Size(2, 2),
			new Size(2, 2),
			new Size(0, 0),
			new Size(0, 0),
			new Size(0, 0),
			new Size(2, 2),
			new Size(2, 2),
			new Size(2, 4),
			new Size(1, 1)
		});


		/// <summary>
		/// Enemy sprites used in Kraid's Hideout.
		/// </summary>
		static public Sprite[] KraidSprites = Sprite.FromArray(new byte[][]{
			new byte[]{SpriteMacros.MoveL4, 0xE2, 0xE3, SpriteMacros.FlipX, 0xE2, SpriteMacros.NextRow,
						SpriteMacros.MoveL4, 0xF2, 0xF3, SpriteMacros.FlipX, 0xF2},
			new byte[]{SpriteMacros.MoveL4, SpriteMacros.FlipY, 0xF2, SpriteMacros.FlipY, 0xF3, SpriteMacros.FlipY, SpriteMacros.FlipX, 0xF2, SpriteMacros.NextRow,
						SpriteMacros.MoveL4, SpriteMacros.FlipY, 0xE2, SpriteMacros.FlipY, 0xE3, SpriteMacros.FlipY, SpriteMacros.FlipX, 0xE2},
			new byte[]{},
			new byte[]{0xC0, 0xC1},
			new byte[]{SpriteMacros.MoveR4,SpriteMacros.MoveR4, 0xCE, SpriteMacros.NextRow,
						SpriteMacros.MoveR4,0xDE, 0xDF, SpriteMacros.NextRow,
						SpriteMacros.MoveR4,0xEE, SpriteMacros.FlipX, 0xEE},
			new byte[]{0xCC, 0xCD, SpriteMacros.NextRow, 0xDC, 0xDD},
			new byte[]{},
			new byte[]{SpriteMacros.MoveL4, SpriteMacros.MoveL4, 0xC2, 0xC3, SpriteMacros.NextRow, SpriteMacros.MoveL4, SpriteMacros.MoveL4, 0xD2, 0xD3},
			new byte[]{0xC5, 0xC6, 0xC7, SpriteMacros.NextRow,
						0xD5, 0xD6, 0xD7, SpriteMacros.NextRow,
						0xE5, 0xE6, 0xE7, SpriteMacros.NextRow,
						0xF5, 0xF6, 0xF7}
		}, new Size[] {
			new Size(2, 2),
			new Size(2, 2),
			new Size(0, 0),
			new Size(2, 1),
			new Size(2, 3),
			new Size(2, 2),
			new Size(0, 0),
			new Size(2, 2),
			new Size(3, 4)
		});
		/// <summary>
		/// Enemy sprites used in Ridley's Hideout.
		/// </summary>
		static public Sprite[] RidleySprites = Sprite.FromArray(new byte[][]{
			new byte[]{SpriteMacros.MoveL4, 0xCE, 0xCF, SpriteMacros.FlipX, 0xCE, SpriteMacros.NextRow,
						SpriteMacros.MoveR4, 0xDE, SpriteMacros.FlipX, 0xDE},
			new byte[]{},
			// Check: these next two might be backwards order
			new byte[]{SpriteMacros.MoveL4, 0xE2, 0xE3, SpriteMacros.FlipX, 0xE2, SpriteMacros.NextRow,
						SpriteMacros.MoveL4, 0xF2, 0xF3, SpriteMacros.FlipX, 0xF2},
			new byte[]{SpriteMacros.MoveL4, SpriteMacros.FlipY, 0xF2, SpriteMacros.FlipY, 0xF3, SpriteMacros.FlipY, SpriteMacros.FlipX, 0xF2, SpriteMacros.NextRow,
						SpriteMacros.MoveL4, SpriteMacros.FlipY, 0xE2, SpriteMacros.FlipY, 0xE3, SpriteMacros.FlipY, SpriteMacros.FlipX, 0xE2},
			new byte[]{},
			new byte[]{},
			new byte[]{0xCC, 0xCD, SpriteMacros.NextRow, 0xDC, 0xDD},
			new byte[]{0xC2, 0xC3, SpriteMacros.NextRow, 0xD2, 0xD3},
			new byte[]{},
			new byte[]{SpriteMacros.MoveU4, 0xC0, 0xC8, 0xC9, SpriteMacros.NextRow,
						SpriteMacros.MoveU4, 0xC0, 0xC6, 0xC7, SpriteMacros.NextRow,
						SpriteMacros.MoveU4, 0xC0, 0xD6, 0xD7, SpriteMacros.NextRow,
						SpriteMacros.MoveU4, 0xE5, 0xE6, 0xE7, SpriteMacros.NextRow,
						SpriteMacros.MoveU4, 0xF5, 0xF6, 0xF7, SpriteMacros.NextRow,
			},
			new byte[]{},
			new byte[]{},
			new byte[]{0xEE, 0xEF, SpriteMacros.NextRow,
						SpriteMacros.FlipX, SpriteMacros.FlipY, 0xEF, SpriteMacros.FlipY, 0xEF},
			new byte[]{0xC8, 0xC9, SpriteMacros.NextRow,
						0xD8, 0xD9, SpriteMacros.NextRow,
						0xE8, 0xE9, SpriteMacros.NextRow,
						0xF8, 0xF9, SpriteMacros.NextRow},
			new byte[]{0XC0}
		}, new Size[] {
			new Size(2, 2),
			new Size(0, 0),
			new Size(2, 2),
			new Size(2, 2),
			new Size(0, 0),
			new Size(0, 0),
			new Size(2, 2),
			new Size(2, 2),
			new Size(0, 0),
			new Size(4, 5),
			new Size(0, 0),
			new Size(0, 0),
			new Size(2, 2),
			new Size(2, 2),
			new Size(2, 2)
		});

		/// <summary>
		/// Enemy sprites used in Tourain.
		/// </summary>
		static public Sprite[] TourainSprites = Sprite.FromArray(new byte[][]{
			new byte[]{0xC0, 0xC1, 0xC2, SpriteMacros.NextRow,
						0xD0, 0xD1, 0xD2, SpriteMacros.NextRow,
						0xE0, 0xE1, 0xE2, SpriteMacros.NextRow},
			new byte[]{0xC3, 0xC4, 0xC5, SpriteMacros.NextRow,
						0xD3, 0xD4, 0xD5, SpriteMacros.NextRow,
						0xE3, 0xE4, 0xE5, SpriteMacros.NextRow}
		}, new Size[]{
			new Size(3, 3),
			new Size(3, 3)
		});
		#endregion

		/// <summary>
		/// Gets sprite definitions based on the specified level.
		/// </summary>
		/// <param name="level">The level to get sprite definitions for.</param>
		/// <returns>An array of sprites.</returns>
		public static Sprite[] GetSprites(Levels level) {
			switch(level) {
				case Levels.Brinstar:
					return BrinstarSprites;
				case Levels.Kraid:
					return KraidSprites;
				case Levels.Ridley:
					return RidleySprites;
				case Levels.Norfair:
					return NorfairSprites;
				case Levels.Tourain:
					return TourainSprites;
				default:
					throw new ArgumentException("Specified level is invalid.", "level");
			}
		}
	}



	/// <summary>
	/// Enumerates special values that can be used in sprite data
	/// </summary>
	static class SpriteMacros
	{
		/// <summary>Tiles drawing will move down one row and back to origional x.
		/// This undoes any Move macros.</summary>
		public const byte NextRow = 0xB0;
		/// <summary>Next tile will be flipped horizontally.</summary>
		public const byte FlipX = 0xB1;
		/// <summary>Next tile will be flipped vertically.</summary>
		public const byte FlipY = 0xB2;
		/// <summary>Tiles will be drawn four pixels to the right.</summary>
		public const byte MoveR4 = 0xB3;
		/// <summary>Tiles will be drawn four pixels to the left.</summary>
		public const byte MoveL4 = 0xB4;
		/// <summary>Tiles will be drawn four pixels up.</summary>
		public const byte MoveU4 = 0xB5;
		/// <summary>Tiles will be drawn four pixels down.</summary>
		public const byte MoveD4 = 0xB6;
		/// <summary>Tile will be rendered with next palette.</summary>
		public const byte NextPal = 0xB7;
	}

}