using System;
using System.Collections.Generic;
using System.Text;
using Editroid.ROM;
using System.Drawing;

namespace Editroid
{
    abstract class ItemEditTool
    {
        public abstract string[] GetList();
        public abstract int GetIndex(ItemSeeker s);
        public abstract void SetIndex(ItemSeeker s, int i);
        public static ItemEditTool GetEditTool(ItemSeeker s) {
            ItemType itemType = (ItemType)(s.Data[s.itemOffset] & 0x0F);
            switch(itemType) {
                case ItemType.Elevator:
                    return elevator;
                case ItemType.Enemy:
                case ItemType.RespawnEnemy:
                    return enemy;
                case ItemType.PowerUp:
                    return pow;
                case ItemType.Nothing:
                    break;
                case ItemType.Door:
                    return door;
                case ItemType.Mella:
                case ItemType.Rinkas:
                case ItemType.PalSwap:
                    return singleByte;
            }

            return unkTool;
        }

        public abstract Image GetImage(ItemSeeker s);

        static EnemyEditTool enemy = new EnemyEditTool();
        static PowerUpEditTool pow = new PowerUpEditTool();
        static ElevatorEditTool elevator = new ElevatorEditTool();
        static SingleByteEditTool singleByte = new SingleByteEditTool();
        static UnknowItemEditTool unkTool = new UnknowItemEditTool();
        static DoorEditTool door = new DoorEditTool();
    }

    class EnemyEditTool:ItemEditTool
    {
        static Image itemImage = ItemImages.Enemy;
        public override Image GetImage(ItemSeeker s) {
            return itemImage;
        }
        static string[] List = { "enemy (0)", "enemy (1)", "enemy (2)", "enemy (3)", "enemy (4)", "enemy (5)", "enemy (6)", "enemy (7)", "enemy (8)", "enemy (9)", "enemy (A)", "enemy (B)", "enemy (C)", "enemy (D)", "enemy (E)", "enemy (F)" };
        public override string[] GetList() {
            return List;
        }

        public override int GetIndex(ItemSeeker s) {
            return s.Data[s.itemOffset + 1] & 0x0F;
        }

        public override void SetIndex(ItemSeeker s, int i) {
            s.Data[s.itemOffset + 1] = (byte)(
                (s.Data[s.itemOffset + 1] & 0xF0) |
                (i & 0x0F));
        }

        public int GetSlot(ItemSeeker s) {
            return s.Data[s.itemOffset] / 0x10;
        }
        public void SetSlot(ItemSeeker s, int i) {
            s.Data[s.itemOffset] = (byte)(
                (s.Data[s.itemOffset] & 0x0F) |
                (i * 0x10));
        }
        public bool GetHard(ItemSeeker s) {
            return s.Data[s.itemOffset + 1] >= 0x80;
        }
        public bool GetRespawn(ItemSeeker s) {
            return (s.Data[s.itemOffset] & 0x07) == 0x07;
        }
        public void SetHard(ItemSeeker s, bool hard) {
            if(hard)
                s.Data[s.itemOffset + 1] = (byte)(s.Data[s.itemOffset + 1] | 0x80);
            else
                s.Data[s.itemOffset + 1] = (byte)(s.Data[s.itemOffset + 1] & 0x7F);
        }
    }
    class PowerUpEditTool:ItemEditTool
    {
        static Image[] itemImages = { ItemImages.Bombs, ItemImages.Boot, ItemImages.Long, ItemImages.Screw, ItemImages.MaruMari, ItemImages.Varia, ItemImages.Wave, ItemImages.Ice, ItemImages.Energy, ItemImages.Missile };
        public override Image GetImage(ItemSeeker s) {
            int i = GetIndex(s);
            if(i < itemImages.Length) return itemImages[i];
            return null;
        }

        static string[] Items = { "Bomb", "High Jump", "Long Beam", "Screw Attack", "Maru Mari", "Varia", "Wave Beam", "Ice Beam", "Energy Tank", "Missile", "(Invalid)", "(Invalid)", "(Invalid)", "(Invalid)", "(Invalid)", "(Invalid)" };
        public override string[] GetList() {
            return Items;
        }

        public override int GetIndex(ItemSeeker s) {
            return s.Data[s.itemOffset + 1] & 0x0F;
        }

        public override void SetIndex(ItemSeeker s, int i) {
            s.Data[s.itemOffset + 1] = (byte)(
                (s.Data[s.itemOffset + 1] & 0xF0) |
                i & 0x0F);
        }
    }
    class ElevatorEditTool:ItemEditTool
    {
        static Image itemImage = ItemImages.Elevator;
        public override Image GetImage(ItemSeeker s) {
            return itemImage;
        }

        static string[] elevatorItems = { "Brinstar", "Norfair", "Kraid", "Tourain", "Ridley", "Norfair Exit", "Kraid Exit", "Tourain Exit", "Ridley Exit", "Complete Game" };

        public override string[] GetList() {
            return elevatorItems;
        }

        public override int GetIndex(ItemSeeker s) {
            int dest = s.Data[s.itemOffset + 1];

            if((dest & 0xF) == 0xF) dest -= 0xA; // Complete game
            if((dest & 0x80) == 0x80) dest += 4; // Exit
            dest = dest & 0x7F; // I forget why
            return dest;
        }

        public override void SetIndex(ItemSeeker s, int i) {
            if(i == 9)
                s.Data[s.itemOffset + 1] = (byte)(ElevatorDestination.GameEnd | ElevatorDestination.ElevatorUp);
            else if(i > 4) {
                i = (i - 4) | 0x80;
            }

            s.Data[s.itemOffset + 1] = (byte)i;
        }
    }
    class SingleByteEditTool:ItemEditTool
    {
        static Image[] itemImage = { ItemImages.Mella, ItemImages.Rinka, ItemImages.Pal };

        public override Image GetImage(ItemSeeker s) {
            return itemImage[GetIndex(s)];
        }
        static string[] SingleByteItems = { "Mella", "Rinka", "Palette Swap" };

        public override string[] GetList() {
            return SingleByteItems;
        }

        public override int GetIndex(ItemSeeker s) {
            ItemType i = (ItemType)(s.Data[s.itemOffset] & 0x0F);
            if(i == ItemType.Mella) return 0;
            if(i == ItemType.Rinkas) return 1;
            return 2;
        }

        public override void SetIndex(ItemSeeker s, int i) {
            ItemType type;
            if(i == 0)
                type = ItemType.Mella;
            else if(i == 1)
                type = ItemType.Rinkas;
            else
                type = ItemType.PalSwap;

            s.Data[s.itemOffset] = (byte)(
                (s.Data[s.itemOffset] & 0xF0) |
                ((int)type & 0x0F));
        }
    }
    class UnknowItemEditTool:ItemEditTool
    {

        static Image UnknownImage = ItemImages.Unknown;
        public override Image GetImage(ItemSeeker s) {
            return UnknownImage;
        }
        public override string[] GetList() {
            return new string[] { "unknown" };
        }

        public override int GetIndex(ItemSeeker s) {
            return 0;
        }

        public override void SetIndex(ItemSeeker s, int i) {
        }
    }
    class DoorEditTool:ItemEditTool
    {

        static Image DoorImage  = ItemImages.Door;
        public override Image GetImage(ItemSeeker s) {
            return DoorImage;
        }
        public override string[] GetList() {
            return new string[] { "Door" };
        }

        public override int GetIndex(ItemSeeker s) {
            return 0;
        }

        public override void SetIndex(ItemSeeker s, int i) {
        }
    }
}
