using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Editroid.Properties;

namespace Editroid
{
    /// <summary>
    /// Allows a user to navigate and edit the map of the world.
    /// </summary>
    public partial class MapControl:PictureBox
    {
        /// <summary>
        /// Instantiates this control.
        /// </summary>
        public MapControl() {
            base.Image = this.Display;
            this.gDisplay = Graphics.FromImage(this.Display);
            this.Cursor = Cursors.Cross;

            for(int i = 0; i < 0x20; i++) {
                for(int j = 0; j < 0x20; j++) {
                    this.MapData[i, j] = MapRoom.Blank;
                }
            }

            this.Selection = new PictureBox();
            this.Selection.BackColor = Color.Transparent;
            ((PictureBox)this.Selection).Image = Resources.Select;
            ((PictureBox)this.Selection).SizeMode = PictureBoxSizeMode.StretchImage;
            this.Selection.Text = null;
            this.Selection.Size = new Size(8, 8);
            this.Selection.Cursor = Cursors.Arrow;
            base.Controls.Add(this.Selection);

            this.Selection.MouseDown += new MouseEventHandler(Selection_MouseDown);
            this.Selection.MouseMove += new MouseEventHandler(Selection_MouseMove);
            this.Selection.MouseUp += new MouseEventHandler(Selection_MouseUp);
        }

        /// <summary>
        /// Gets/sets the size of the map selection.
        /// </summary>
        public Size SelectionSize {
            get {
                return new Size(Selection.Width / 8, Selection.Height / 8);
            }
            set {
                Selection.Size = new Size(value.Width * 8, value.Height * 8);
            }
        }


        bool isDraggingSelection;
        int selectionDragX, selectionDragY;

        void Selection_MouseUp(object sender, MouseEventArgs e) {
            //OnMouseUp(new MouseEventArgs(e.Button, e.Clicks, e.X + Selection.Left, e.Y + Selection.Top, e.Delta));
            isDraggingSelection = false;
        }

        void Selection_MouseMove(object sender, MouseEventArgs e) {
            //OnMouseMove(new MouseEventArgs(e.Button, e.Clicks, e.X + Selection.Left, e.Y + Selection.Top, e.Delta));
            if(isDraggingSelection) {
                int dx = e.X / 8 - selectionDragX;
                int dy = e.Y / 8 - selectionDragY;

                if(e.X < selectionDragX) dx -= 8;
                if(e.Y < selectionDragY) dy -= 8;
                //if(dy < 0) dy -= 8;

                //dx /= 8;
                //dy /= 8;

                if(dx != 0 || dy != 0)
                    SetSelection(SelectionX + dx, SelectionY + dy);
            }
        }

        void Selection_MouseDown(object sender, MouseEventArgs e) {
            //OnMouseDown(new MouseEventArgs(e.Button, e.Clicks, e.X + Selection.Left, e.Y + Selection.Top, e.Delta));
            if(e.Button == MouseButtons.Left) {
                isDraggingSelection = true;
                selectionDragX = e.X / 8;
                selectionDragY = e.Y / 8;
            }
        }

        /// <summary>
        /// Gets door data for a specified map position.
        /// </summary>
        /// <param name="x">X-coordinate on the map.</param>
        /// <param name="y">Y-coordinate on the map.</param>
        /// <returns>MapRoom value indicating which doors are present.</returns>
        /// <remarks>This function gets data that is displayed, which does not necessarily
        /// accurately reflect the doors in the game.</remarks>
        public MapRoom GetDoors(int x, int y) {
            return this.MapData[x, y] & (MapRoom.LeftDoor | MapRoom.RightDoor);
        }

        /// <summary>
        /// Gets level data for the currently selected map position.
        /// </summary>
        /// <returns>A Levels value.</returns>
        public Levels GetLevel() {
            return GetLevel(SelectionX, SelectionY);
        }

        /// <summary>
        /// Gets/sets the X-coordinate of the selected screen.
        /// </summary>
        public int SelectionX {
            get {
                return Selection.Left / 8;
            }
            set {
                if(value < 0) value = 0;
                if(value > 312) value = 31;

                Selection.Left = value * 8;
                if(SelectionChanged != null) SelectionChanged(this, value, SelectionY);
                Application.DoEvents();
            }
        }


        /// <summary>
        /// Moves the map selection to the specified coordinates.
        /// </summary>
        /// <param name="x">X coordinate.</param>
        /// <param name="y">Y coordinate.</param>
        public void SetSelection(int x, int y) {
            if(x < 0) x = 0; if (x > 31) x = 31;
            if(y < 0) y = 0; if (y > 31) y = 31;
            Selection.Location = new Point(x * 8, y * 8);
            if(SelectionChanged != null) SelectionChanged(this, x, y);
        }


        /// <summary>
        /// Gets/sets the Y-coordinate of the selected screen.
        /// </summary>
        public int SelectionY {
            get {
                return Selection.Top / 8;
            }
            set {
                if(value < 0) value = 0;
                if(value > 312) value = 31;

                Selection.Top = value * 8;
                if(SelectionChanged != null) SelectionChanged(this, SelectionX, value);
                Application.DoEvents();
            }
        }

        /// <summary>
        /// Gets level data for a specified map position.
        /// </summary>
        /// <param name="x">X-coordinate of a screen.</param>
        /// <param name="y">Y-coordinate of a screen.</param>
        /// <returns>A Levels value.</returns>
        /// <remarks>This function gets data that is displayed and does not
        /// necessarily accurately reflect the game world.</remarks>
        public Levels GetLevel(int x, int y) {
            MapRoom level = this.MapData[x, y] & ((MapRoom)0xfc);
            if(level == MapRoom.Ridley)
                return Levels.Ridley;
            else if(level == MapRoom.Kraid)
                return Levels.Kraid;
            else if(level == MapRoom.Tourain)
                return Levels.Tourain;
            else if(level == MapRoom.Norfair)
                return Levels.Norfair;
            else if(level == MapRoom.Blank)
                return Levels.Invalid;
            else
                return Levels.Brinstar;
        }


        /// <summary>
        /// Gets map data.
        /// </summary>
        /// <param name="x">X-coordinate of a screen.</param>
        /// <param name="y">Y-coordinate of a screen.</param>
        /// <returns>A MapRoom value describing a map position.</returns>
        /// <remarks>This function gets data that is displayed and does not
        /// necessarily accurately reflect the game world.</remarks>
        public MapRoom GetRoom(int x, int y) {
            return this.MapData[x, y];
        }

        /// <summary>
        /// Indicates whether the control is currently loading data, so that
        /// events that should not occur at this time will be surpressed.
        /// </summary>
        bool loading;

        /// <summary>
        /// Loads a map from a stream.
        /// </summary>
        /// <param name="s">Stream containing map data.</param>
        public bool TryLoadData(Stream s) {
            return TryLoadData(s, true);
        }
        /// <summary>
        /// Loads a map from a stream.
        /// </summary>
        /// <param name="s">Stream containing map data.</param>
        /// <param name="hasHeader"></param>
        /// <returns>A boolean value that indicates success or failure.</returns>
        public bool TryLoadData(Stream s, bool hasHeader) {
            loading = true;
            try {
                if(hasHeader) {
                    for(int i = 0; i < MagicValue.Length; i++) {
                        try {
                            if(s.CanRead == false)
                                return false;

                            if(s.ReadByte() != MagicValue[i])
                                return false;
                        }
                        catch(Exception) {
                            return false;
                        }
                    }

                    Invalidate();
                }
                for(int i = 0; i < 0x20; i++) {
                    for(int j = 0; j < 0x20; j++) {
                        this.SetImage(j, i, (MapRoom)((byte)s.ReadByte()));
                    }
                }

                return true;
            }
            finally {
                loading = false;
            }
        }

        /// <summary>
        /// Loads a map from a file.
        /// </summary>
        /// <param name="filename">Name of file containing map data.</param>
        public void LoadData(string filename) {
            using(FileStream s = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)) {
                this.TryLoadData(s, true);
            }
            base.Invalidate();
        }


        /// <summary>
        /// Overrides base member.
        /// </summary>
        /// <param name="e">e parameter.</param>
        protected override void OnMouseDown(MouseEventArgs e) {
            base.OnMouseDown(e);
            int x = e.X / 8;
            int y = e.Y / 8;
            x = (x < 0) ? 0 : x;
            y = (y < 0) ? 0 : y;
            x = (x > 0x1f) ? 0x1f : x;
            y = (y > 0x1f) ? 0x1f : y;
            if((Control.ModifierKeys & Keys.Shift) == Keys.Shift) {

                // Shift-Left-Click
                if(e.Button == MouseButtons.Left) {
                    // If clicked screen is not blank, begin drawing on the map with
                    // the clicked color
                    if(MapData[x, y] != MapRoom.Blank) {
                        this.Drawing = true;
                        this.pen = this.MapData[x, y] & ~(MapRoom.LeftDoor | MapRoom.RightDoor);
                    }

                    // Shift-Right-Click
                } else if(e.Button == MouseButtons.Right) {
                    // Cycle to the next level.
                    MapRoom room = this.MapData[x, y];
                    if(room == MapRoom.Blank || room == MapRoom.Ridley) {
                        // Cylce from ridley to brinstar (skip blank)
                        room = MapRoom.Brinstar;
                    } else {
                        // Go to next level in cycle
                        room = (MapRoom)(room + 4); // 0x1 and 0x2 are flags, so room values are 4 apart
                    }
                    this.SetImage(x, y, room);
                    if(this.TileChanged != null) {
                        this.TileChanged(this, x, y);
                    }
                    base.Invalidate();
                }
            } else {

                // Right Click
                if(e.Button == MouseButtons.Right) {
                    //this.Selection.Location = new Point(x * 8, y* 8);

                    // Clear the room
                    this.SetImage(x, y, MapRoom.Blank);
                    Invalidate();
                    if(this.RoomCleared != null)
                        RoomCleared(this, x, y);
                
                // Left Click
                } else{
                    int mapX = e.X - (SelectionSize.Width - 1) * 4;
                    int mapY = e.Y - (SelectionSize.Height - 1) * 4;
                    mapX /= 8; mapY /= 8;
                    if(mapX < 0) mapX = 0; if(mapX > 31) mapX = 31;
                    if(mapY < 0) mapY = 0; if(mapY > 31) mapY = 31;

                    this.Selection.Location = new Point(mapX * 8, mapY * 8);
                    
                    // Select the clicked screen
                     if(this.SelectionChanged != null) 
                         this.SelectionChanged(this, mapX, mapY);
                }
            }
            
        }

        /// <summary>
        /// Overrides base member.
        /// </summary>
        /// <param name="e">e parameter.</param>
        protected override void OnMouseMove(MouseEventArgs e) {
            base.OnMouseMove(e);
            if(this.Drawing) {
                int x = e.X / 8;
                int y = e.Y / 8;
                x = (x < 0) ? 0 : x;
                y = (y < 0) ? 0 : y;
                x = (x > 0x1f) ? 0x1f : x;
                y = (y > 0x1f) ? 0x1f : y;
                if(this.MapData[x, y] != this.pen) {
                    this.SetImage(x, y, this.pen);
                    if(this.TileChanged != null) {
                        this.TileChanged(this, x, y);
                    }
                    base.Invalidate();
                }
            }
        }

        /// <summary>
        /// Overrides base member.
        /// </summary>
        /// <param name="e">e parameter.</param>
        protected override void OnMouseUp(MouseEventArgs e) {
            base.OnMouseUp(e);
            if(e.Button == MouseButtons.Left) {
                this.Drawing = false;
            }
        }


        /// <summary>Binary data marker to identify map data.</summary>
        static byte[] MagicValue = { 0x45, 0x44, 0x49, 0x54, 0x52, 0x4F, 0x49, 0x44 };
        /// <summary>
        /// Saves map to a stream.
        /// </summary>
        /// <param name="s">Stream to write data to.</param>
        public void SaveData(Stream s) {
            s.Write(MagicValue, 0, MagicValue.Length);

            for(int i = 0; i < 0x20; i++) {
                for(int j = 0; j < 0x20; j++) {
                    s.WriteByte((byte)this.MapData[j, i]);
                }
            }
        }


        /// <summary>
        /// Saves map to a file.
        /// </summary>
        /// <param name="filename">Filename.</param>
        public void SaveData(string filename) {
            using(FileStream s = new FileStream(filename, FileMode.OpenOrCreate)) {
                this.SaveData(s);
            }
        }




        /// <summary>Raised when the selected room is modified.</summary>
        public event EventHandler selectedRoomModified;

        /// <summary>
        /// Sets the image to be used at a specified map position.
        /// </summary>
        /// <param name="x">Coordinate of screen.</param>
        /// <param name="y">Coordinate of screen.</param>
        /// <param name="image">Image to display.</param>
        public void SetImage(int x, int y, MapRoom image) {
            if(image > MapRoom.Blank) {
                image = MapRoom.Blank;
            }
            if(this.MapData[x, y] != image) {
                this.bltDest.X = x * 8;
                this.bltDest.Y = y * 8;
                this.bltSource.X = (int)image * 8;
                this.gDisplay.DrawImage(this.Source, this.bltDest, this.bltSource, GraphicsUnit.Pixel);
                this.MapData[x, y] = image;
                if(image != MapRoom.Blank && rom != null && rom.GetRoomIndex(x, y) == 0xFF) {
                }

                if(!loading && 
                    SelectionX == x && SelectionY == y
                    && selectedRoomModified != null)

                    selectedRoomModified(this, new EventArgs());

                if(!loading && RoomSet != null) {
                    RoomSet(this, x, y);
                }
            }
        }


        private bool ShouldSerializeImage() {
            return false;
        }
        /// <summary>
        /// Hides the Image property.
        /// </summary>
        [EditorBrowsable(EditorBrowsableState.Never), Browsable(false)]
        public new Image Image {
            get {
                return base.Image;
            }
            set {
            }
        }






        // Nested Types
        /// <summary>
        /// Represents an event with coordinate data.
        /// </summary>
        /// <param name="sender">Object that raises the event.</param>
        /// <param name="X">X coordinate.</param>
        /// <param name="Y">Y coordinate.</param>
        public delegate void CoordEvent(object sender, int X, int Y);

        // Events
        /// <summary>
        /// Occurs when a room is selected.
        /// </summary>
        public event CoordEvent SelectionChanged;
        /// <summary>
        /// Occurs when a room is cleared.
        /// </summary>
        public event CoordEvent RoomCleared;
        /// <summary>
        /// Occurs when a room is changed.
        /// </summary>
        public event CoordEvent RoomSet;
        /// <summary>
        /// Occurs when a map tile is changed.
        /// </summary>
        public event CoordEvent TileChanged;

        // Fields
        private Bitmap Display = (Bitmap)Resources.MapBase.Clone();
        private Graphics gDisplay;
        private Bitmap Source = Resources.MapTiles;

        private Control Selection;

        private bool Drawing = false;
        private Rectangle bltDest = new Rectangle(0, 0, 8, 8);
        private Rectangle bltSource = new Rectangle(0, 0, 8, 8);

        private MapRoom[,] MapData = new MapRoom[0x20, 0x20];
        private MapRoom pen = MapRoom.Blank;

        /// <summary>
        /// Gets the image currently being used for the map.
        /// </summary>
        public Bitmap MapImage { get { return Display; } }
        /// <summary>
        /// Displays the specified doors on the specified room in the map
        /// </summary>
        /// <param name="X">The X-coordinate of the room</param>
        /// <param name="Y">The Y-coordinate of the room</param>
        /// <param name="doors">The door data to apply</param>
        public void SetDoor(int X, int Y, MapRoom doors) {
            this.SetImage(X, Y, (MapData[X, Y] & ~(MapRoom.LeftDoor | MapRoom.RightDoor)) | doors);
        }

        internal void SetSelectionLevel(MapRoom level) {
            SetImage(SelectionX, SelectionY, level);
            Selection.Invalidate();
        }

        internal void Filter(Rom GameRom) {
            for(int x = 0; x < 32; x++) {
                for(int y = 0; y < 32; y++) {
                    if(GameRom.GetRoomIndex(x, y) == 0xFF)
                        SetImage(x, y, MapRoom.Blank);
                }
            }
        }

        private Rom rom;

        /// <summary>
        /// Gets/sets the ROM.
        /// </summary>
        public Rom Rom {
            get { return rom; }
            set { rom = value; }
        }

    }


	

    /// <summary>
    /// Enumerates values that can be combined to specifiy the image to use
    /// for a screen on a map.
    /// </summary>
    [Flags]
    public enum MapRoom:byte
    {
        /// <summary>Screen is in Brinstar.</summary>
        Brinstar = 0,
        /// <summary>Screen is in Norfair.</summary>
        Norfair = 4,
        /// <summary>Screen is in Tourain.</summary>
        Tourain = 8,
        /// <summary>Screen is in Kraid's hideout.</summary>
        Kraid = 0xC,
        /// <summary>Screen is blank.</summary>
        Blank = 0x14,
        /// <summary>Screen is in Ridley's hideout.</summary>
        Ridley = 0x10,
        /// <summary>Screen has a door on the left side.</summary>
        LeftDoor = 1,
        /// <summary>Screen has a door on the right side.</summary>
        RightDoor = 2
    }


}
