using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using Editroid.Graphic;

namespace Editroid
{
    /// <summary>
    /// Creates a map 
    /// </summary>
    internal class MapMaker
    {

        public const int ScreenWidth = 256;
        public const int ScreenHeight = 240;

        private Rom rom;

        public Rom Rom {
            get { return rom; }
            set { rom = value; }
        }

        private bool showPhysics;

        public bool ShowPhysics {
            get { return showPhysics; }
            set { showPhysics = value; }
        }

        private bool fillEmptySpots;

        public bool FillEmptySpots {
            get { return fillEmptySpots; }
            set { fillEmptySpots = value; }
        }

        private Rectangle mapArea;

        public Rectangle MapArea {
            get { return mapArea; }
            set { mapArea = value; }
        }

        private Levels levelFilter = Levels.Invalid;

        public Levels LevelFilter {
            get { return levelFilter; }
            set { levelFilter = value; }
        }

        private int scale = 1;

        public int Scale {
            get { return scale; }
            set { scale = value; }
        }

        private MapControl levelMap;

        public MapControl LevelMap {
            get { return levelMap; }
            set { levelMap = value; }
        }

        private bool highQuality;

        public bool HighQuality {
            get { return highQuality; }
            set { highQuality = value; }
        }

        private bool hideEnemies;

        public bool HideEnemies {
            get { return hideEnemies; }
            set { hideEnemies = value; }
        }


        ScreenDisplay renderer = new ScreenDisplay();
        Blitter b = new Blitter();
        Bitmap screenBitmap = new Bitmap(ScreenWidth, ScreenHeight, PixelFormat.Format8bppIndexed);

        Bitmap map;
        Graphics gMap;

        public Bitmap Render() {
            map = new Bitmap(mapArea.Width * ScreenWidth / scale, mapArea.Height * ScreenHeight / scale, PixelFormat.Format24bppRgb);
            gMap = Graphics.FromImage(map);
            gMap.InterpolationMode = highQuality ?
                System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic :
                System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
            gMap.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Half;


            for(int x = 0; x < mapArea.Width; x++) {
                for(int y = 0; y < mapArea.Height; y++) {
                    DrawScreen(x + mapArea.Left, y + mapArea.Y, x ,y);

                    if(RoomDrawn != null)
                        RoomDrawn(this, new EventArgs());
                }
            }

            return map;
        }

        private void DrawScreen(int gameX, int gameY, int mapx, int mapY) {
            // Get level and screen
            Levels level = levelMap.GetLevel(gameX, gameY);
            bool invalidLevel = false;

            if(level == Levels.Invalid ||
                (levelFilter != Levels.Invalid && level != levelFilter)) {

                if(!fillEmptySpots) return;

                if(levelFilter != Levels.Invalid)
                    level = levelFilter;
                else 
                    level = GuessLevel(gameX, gameY);

                if(level == Levels.Tourain)
                    level = Levels.Kraid;

                invalidLevel = true;
            }

            Level levelData = rom.GetLevel(level);
            int screenIndex = rom.GetRoomIndex(gameX, gameY);
            if(screenIndex == 0xFF || invalidLevel) { // Blank screen
                screenIndex = GetBlankScreenIndex(level);
            }


            Screen screen = levelData.GetScreen(screenIndex);
            // Load data into renderer
            renderer.Level = levelData;
            renderer.SelectedEnemy = -1;

            // Apply palette
            ColorPalette bitmapPalette = screenBitmap.Palette; // Get bitmap palette
            
            NesPalette bgPalette = levelData.BgPalette; // Get ROM palettes
            NesPalette spritePalette = levelData.SpritePalette;
            bgPalette.ApplyTable(bitmapPalette.Entries); // Apply to bitmap palette
            bgPalette.ApplyTable(bitmapPalette.Entries, 16);
            spritePalette.ApplyTable(bitmapPalette.Entries, 32);
            spritePalette.ApplyTable(bitmapPalette.Entries, 48);
            
            screenBitmap.Palette = bitmapPalette; // Apply back to bitmap palette

            // Render Screen
            renderer.DefaultPalette = screen.ColorAttributeTable;
            renderer.Clear();
            renderer.DrawScreen(screenIndex);
            ScreenEnemy[] enemies = hideEnemies ? null : screen.Enemies;
            renderer.Render(b, screenBitmap, enemies, screen.Doors, showPhysics);

            Rectangle source = new Rectangle(0, 0, ScreenWidth, ScreenHeight);
            Rectangle dest = new Rectangle(
                mapx * ScreenWidth / scale, mapY * ScreenHeight / scale,
                ScreenWidth / scale, ScreenHeight / scale);

            gMap.DrawImage(screenBitmap, dest, source, GraphicsUnit.Pixel);

        }

        private int GetBlankScreenIndex(Levels level) {
            switch(level) {
                case Levels.Brinstar:
                    return 0x08;
                case Levels.Norfair:
                    return 0x06;
                case Levels.Tourain:
                    return 0x0B;
                case Levels.Kraid:
                    return 0x0B;
                case Levels.Ridley:
                    return 0x05;
            }

            return 0x08;
        }

        private Levels GuessLevel(int x, int y) {
            Levels l;
            l = levelAt(x, y - 1);
            if(l != Levels.Invalid) return l;
            l = levelAt(x, y + 1);
            if(l != Levels.Invalid) return l;
            l = levelAt(x - 1, y);
            if(l != Levels.Invalid) return l;
            l = levelAt(x + 1, y);
            if(l != Levels.Invalid) return l;
            l = levelAt(x - 1, y - 1);
            if(l != Levels.Invalid) return l;
            l = levelAt(x + 1, y - 1);
            if(l != Levels.Invalid) return l;
            l = levelAt(x - 1, y + 1);
            if(l != Levels.Invalid) return l;
            l = levelAt(x + 1, y + 1);
            if(l != Levels.Invalid) return l;

            return Levels.Brinstar;
        }

        /// <summary>For use exclusively by GuessLevel as a safe
        /// way to check the level at a map position.</summary>
        private Levels levelAt(int x, int y) {
            if(x < 0 || y < 0 || x > 31 || y > 31) return Levels.Invalid;
            return levelMap.GetLevel(x, y);
        }
        /// <summary>
        /// Raised each time a room is drawn to the map. This event will also be
        /// raised when the renderer skips over a blank room.
        /// </summary>
        public event EventHandler RoomDrawn;

    }
}
