using System;
using System.Collections.Generic;
using System.Text;

namespace Editroid
{

    /// <summary>
    /// Represents a table of combos from which structs are made of.
    /// </summary>
    public class ComboTable
    {
        byte[] _Data;
        int _Offset;

        /// <summary>
        /// Gets the offset of this object in ROM data.
        /// </summary>
        public int Offset { get { return _Offset; } }


        /// <summary>
        /// Creates a ComboTable based on the specified ROM data
        /// </summary>
        /// <param name="data">ROM image</param>
        /// <param name="offset">location of data</param>
        public ComboTable(byte[] data, int offset) {
            _Data = data;
            _Offset = offset;
        }

        /// <summary>
        /// Gets a reference to the specified combo definition.
        /// </summary>
        /// <param name="i">Index of a combo.</param>
        /// <returns>A reference to the specified combo defintion.</returns>
        public Combo this[int i] {
            get {
                //Todo: validate i is 0 to 255 and combo is within ROM
                return new Combo(_Data, _Offset + i * 4);
            }
        }
    }

    /// <summary>Represents a combo.</summary>
    public struct Combo
    {
        byte[] _Data;
        int _Offset;

        /// <summary>
        /// Creates a Combo object based on ROM data
        /// </summary>
        /// <param name="data">ROM image</param>
        /// <param name="offset">location of data</param>
        public Combo(byte[] data, int offset) {
            _Data = data;
            _Offset = offset;
        }

        /// <summary>
        /// Gets the index of a specified tile (0 through 3), performing saftey checks to make
        /// sure that the index is within range.
        /// </summary>
        /// <param name="i">tile index</param>
        /// <returns>index of graphic to use for a specific tile</returns>
        public byte this[int i] {
            get {
                if(i < 0 || i > 3) throw new ArgumentException("index was out of bounds.");
                if(_Offset + i >= _Data.Length) throw new Exception("array to small, data not present");

                return _Data[_Offset + i];
            }
            set {
                if(i < 0 || i > 3) throw new ArgumentException("index was out of bounds.");
                if(_Offset + i >= _Data.Length) throw new Exception("array to small, data not present");
                _Data[_Offset + i] = value;
            }
        }

        /// <summary>
        /// Loads a combo tile index. This function performs no safety checks.
        /// </summary>
        /// <param name="index">The tile index. Should be 0, 1, 2, or 3.</param>
        /// <returns>A combo tile index</returns>
        public byte GetByte(int index) {
            return _Data[_Offset + index];
        }

        /// <summary>
        /// Sets a combo tile index. This function performs no saftey checks.
        /// </summary>
        /// <param name="index">The tileindex. Should be 0, 1, 2, or 3.</param>
        /// <param name="value">The new combo tile index.</param>
        public void SetByte(int index, byte value) {
            _Data[_Offset + index] = value;
        }

        /// <summary>
        /// The offset of combo data
        /// </summary>
        public static int BrinstarComboOffset {
            get {
                return 0x6F00;
            }
        }
        /// <summary>
        /// The offset of combo data
        /// </summary>
        public static int NorfairComboOffset {
            get {
                return 0xAEFC;
            }
        }
        /// <summary>
        /// The offset of combo data
        /// </summary>
        public static int KraidComboOffset {
            get {
                //return 0x16C42;
                return 0x12C42;
            }
        }
        /// <summary>
        /// The offset of combo data
        /// </summary>
        public static int RidleyComboOffset {
            get {
                //return 0x12C42;
                return 0x16B33;
            }
        }
        /// <summary>
        /// The offset of combo data
        /// </summary>
        public static int TourainComboOffset {
            get {
                return 0x12E59;
            }
        }

        /// <summary>
        /// Gets the combo offset of a specified level.
        /// </summary>
        /// <param name="level">The level to get combo offsets for.</param>
        /// <returns>An integer representing a data offset.</returns>
        public static int GetComboOffset(Levels level) {
            switch(level) {
                case Levels.Kraid:
                    return KraidComboOffset;
                case Levels.Brinstar:
                    return BrinstarComboOffset;
                case Levels.Ridley:
                    return RidleyComboOffset;
                case Levels.Tourain:
                    return TourainComboOffset;
                case Levels.Norfair:
                    return NorfairComboOffset;
                default:
                    throw new ArgumentException("An invalid level identifier was specified when retrieving combo data from ComboTable.GetComboOffset().", "level");
            }
        }
    }
}
