using System;
using System.Collections.Generic;
using System.Text;

namespace Editroid.ROM
{
    /// <summary>
    /// Represents item data for a level.
    /// </summary>
    public class ItemTable: IEnumerable<ItemEntry>
    {
        Levels level;
        byte[] data;
        List<ItemEntry> entries = new List<ItemEntry>();

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="level"></param>
        /// <param name="data"></param>
        public ItemTable(Levels level, byte[] data) {
            this.level = level;
            this.data = data;

            LoadEntries();
        }

        private void LoadEntries() {
            int offset = GetItemTableOffset(level, data);
            entries.Add(new ItemEntry(level, data, offset));


            while(entries[entries.Count - 1].HasNextEntry) {
                entries.Add(new ItemEntry(level, data, 
                    entries[entries.Count - 1].NextEntryOffset));
            }
        }

        /// <summary>
        /// Re-orders item data entries to prevent parsing problems caused by
        /// entries being out of logical order.
        /// </summary>
        public void ResortEntries() {
            if(entries.Count < 2) return; // No need for sorting.

            //int romDataOffset = GetItemTableOffset(level);
            //int workingOffset = entries[0].NextEntryPointer.RealValue - entries[1].NextEntryOffset;

            // Use a bubble sorting method. SIMPLEST
            bool done = false;
            while(!done) {
                done = true; // Assume we are done unless we find we need more sorting

                for(int i = 0; i < entries.Count - 1; i++) { 
                    if(entries[i].MapY > entries[i + 1].MapY) { // If two are out of order
                        done = false; // Make at least one more sweep

                        // Swap pointers
                        
                        // go from  X -> A -> B -> Y
                        //
                        // to        /--------v
                        //          X    A <- B    Y
                        //                \--------^
                        byte swap1, swap2;
                        int offsetA = entries[i].Offset;
                        int offsetB = entries[i + 1].Offset;
                        int offsetX;
                        if(i > 0) {// if A is not the first.
                            offsetX = entries[i - 1].Offset; // Gets the offset of previous entry
                        } else { // A is the first entry
                            // Use a pointer to the pointer to the item table and treat that like an entry
                            // so we can use the same code
                            MetroidPointer itemTablePointerOffset = GetItemTablePointerOffset();
                            offsetX = itemTablePointerOffset.GetDataOffset(level) - 1;
                        }

                        // Store X.Next(old)
                        swap1 = data[offsetX + 1];
                        swap2 = data[offsetX + 2];

                        // Set X.Next(new) to A.Next(old)
                        data[offsetX + 1] = data[offsetA + 1];
                        data[offsetX + 2] = data[offsetA + 2];

                        // Set A.Next(new) to B.Next(old)
                        data[offsetA + 1] = data[offsetB + 1];
                        data[offsetA + 2] = data[offsetB + 2];

                        // Set B.Next(new) to X.Next(old);
                        data[offsetB + 1] = swap1;
                        data[offsetB + 2] = swap2;

                        // Reflect resorted entries:
                        ItemEntry swapE = entries[i];
                        entries[i] = entries[i + 1];
                        entries[i + 1] = swapE;

                    }
                }
            }
        }


        /// <summary>
        /// Gets the offset of an item table.
        /// </summary>
        /// <param name="level">The level the item table belongs to.</param>
        /// <param name="data">ROM data to obtain offset from.</param>
        /// <returns>The offset of an item table.</returns>
        public static int GetItemTableOffset(Levels level, byte[] data) {
            // Gets a pointer to the pointer to the table offset.
            int pointerOffset = GetItemTablePointerOffset().GetDataOffset(level);
            // Resolves to a pointer to the table offset
            MetroidPointer tablePointer = new MetroidPointer(data, pointerOffset);
            // Resolves to the table offset
            return tablePointer.GetDataOffset(level);
            
            
            //switch(level) {
            //    case Levels.Brinstar:
            //        return 0x63E6;
            //    case Levels.Norfair:
            //        return 0xA2E9;
            //    case Levels.Tourain:
            //        return 0xE84B;
            //    case Levels.Kraid:
            //        return 0x1227D;
            //    case Levels.Ridley:
            //        return 0x1621D;
            //    default:
            //        return 0x0000;
            //}
        }

        /// <summary>
        /// Gets a offset of a pointer to the item table (or a "handle" to the item table).
        /// </summary>
        /// <returns>A offset of a pointer to the item table (or a "handle" to the item table).</returns>
        public static MetroidPointer GetItemTablePointerOffset() {
            return new MetroidPointer(0x98, 0x95);
        }

        IEnumerator<ItemEntry> IEnumerable<ItemEntry>.GetEnumerator() {
            return entries.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
            return entries.GetEnumerator();
        }
    }
    /// <summary>
    /// Represents an item data entry, which contains one or more items from the
    /// same level on the same horizontal row on the map.
    /// </summary>
    public class ItemEntry
    {
        int offset;
        byte[] data;
        Levels level;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="level"></param>
        /// <param name="data"></param>
        /// <param name="offset"></param>
        public ItemEntry(Levels level, byte[] data, int offset) {
            this.offset = offset;
            this.data = data;
            this.level = level;
        }

        /// <summary>Gets the offset of this entry.</summary>
        public int Offset { get { return offset; } }

        /// <summary>
        /// Gets or sets the vertical map position of this item.
        /// </summary>
        public int MapY {
            get {
                return data[offset];
            }
            set {
                data[offset] = (byte)value;
            }
        }

        /// <summary>
        /// Gets a metroid pointer to the next entry.
        /// </summary>
        public MetroidPointer NextEntryPointer {
            get {
                return new MetroidPointer(data, offset + 1);
            }
        }

        /// <summary>
        /// Gets the offset of the next entry.
        /// </summary>
        public int NextEntryOffset {
            get {
                return NextEntryPointer.GetDataOffset(level);
            }
        }

        /// <summary>Gets whether there is an item entry following this one.</summary>
        public bool HasNextEntry {
            get {
                return data[offset + 1] != 0xFF || data[offset + 2] != 0xFF;
            }
        }

        /// <summary>
        /// Provides an object to enumerate items.
        /// </summary>
        /// <returns>An ItemSeeker.</returns>
        public ItemSeeker Seek() {
            return new ItemSeeker(offset + 3, data);
        }
    }

    /// <summary>
    /// Provides a mechanism to seek through screen entries in an item data entry,
    /// and thorugh items if there are multiple items in a screen.
    /// </summary>
    public struct ItemSeeker
    {
        /// <summary></summary>
        public int startOffset;
        /// <summary></summary>
        public int screenOffset;
        /// <summary></summary>
        public int itemOffset;
        byte[] data;

        /// <summary>Gets the ROM data this object represents.</summary>
        public byte[] Data { get { return data; } }
        /// <summary>
        /// Instantiates this object.
        /// </summary>
        /// <param name="offset">The offset of screen data from an item entry (follows 3-byte header of item data entry).</param>
        /// <param name="data">ROM data.</param>
        public ItemSeeker(int offset, byte[] data) {
            this.data = data;
            this.startOffset = this.screenOffset = offset;
            itemOffset = offset + 2;
        }

        /// <summary>
        /// Gets/sets the x-coordinate of the screen being examined.
        /// </summary>
        public int MapX {
            get {
                return data[screenOffset];
            }
            set {
                data[screenOffset] = (byte)value;
            }
        }

        /// <summary>
        /// Gets the sprite slot. This value is irrelevant unless this data represents an enemy.
        /// </summary>
        public int SpriteSlot {
            get {
                return (data[itemOffset] & 0xF0) / 0x10; // Get upper nibble value.
            }
            set {
                byte bval = (byte)(value & 0x0F * 0x10); // Convert to upper nibble
                data[itemOffset] =(byte)((data[itemOffset] & 0x0F) | bval); // Replace upper nibble of ROM data.
            }
        }

        /// <summary>
        /// Gets/sets what type of item is being examined.
        /// </summary>
        public ItemType ItemType {
            get {
                return (ItemType)(
                    data[itemOffset] & 0x0F); // Return lower nibble
            }
            set {
                byte bval = (byte)value;
                data[itemOffset] = (byte)(
                    (data[itemOffset] & 0xF0) | bval); // Apply to lower nibble
            }
        }

        /// <summary>
        /// Gets the type of powerup being examined, if the current item is a powerup.
        /// </summary>
        public PowerUpType PowerUp {
            get {
                return (PowerUpType)data[itemOffset + 1];
            }
            set {
                data[itemOffset + 1] = (byte)value;
            }
        }

        /// <summary>
        /// Gets/sets the destination of the elevator being examined, if the current item is an elevator.
        /// </summary>
        public ElevatorDestination Destination {
            get {
                return (ElevatorDestination)data[itemOffset + 1];
            }
            set {
                data[itemOffset + 1] = (byte)value;
            }
        }

        /// <summary>
        /// Gets/sets the location of this power-up, if the item being examined is a power-up.
        /// </summary>
        public ScreenCoordinate ScreenPosition {
            get {
                return new ScreenCoordinate(data[itemOffset + 2]);
            }
            set {
                data[itemOffset + 2] = value.Value;
            }
        }

        /// <summary>
        /// Gets the size, in bytes, of the data for the current item.
        /// </summary>
        public int currentItemDataSize {
            get {
                switch(ItemType) {
                    case ItemType.Enemy:
                    case ItemType.RespawnEnemy:
                    case ItemType.PowerUp:
                        return 3;
                    case ItemType.Mella:
                    case ItemType.Rinkas:
                    case ItemType.PalSwap:
                        return 1;
                    case ItemType.Elevator:
                    case ItemType.Door:
                        return 2;
                    case ItemType.Nothing:
                    case ItemType.Unknown_5:
                    case ItemType.AccessBridge:
                    case ItemType.Unknown_b:
                    case ItemType.Unknown_c:
                    case ItemType.Unknown_d:
                    case ItemType.Unknown_e:
                    case ItemType.Unknown_f:
                    default:
                        return 1;
                }
            }
        }

        /// <summary>
        /// Returns true if there are more items in the screen being examined.
        /// </summary>
        public bool MoreItemsPresent {
            get { return data[itemOffset + currentItemDataSize] != 0x00; }
        }

        /// <summary>
        /// Seeks to the next item, if there are more items in the screen.
        /// </summary>
        public void NextItem() {
            itemOffset += 3;
        }
        /// <summary>
        /// Seeks to the first item in the screen being examined.
        /// </summary>
        public void ResetItemPointer(){
            itemOffset = screenOffset + 2;
        }

        /// <summary>
        /// Seeks to the first room in the item entry.
        /// </summary>
        public void ResetSeeker() {
            screenOffset = startOffset;
            itemOffset = startOffset + 2;
        }

        /// <summary>
        /// Gets true if there are more screens in the entry.
        /// </summary>
        public bool MoreScreens {
            get {
                return data[screenOffset + 1]  != 0xFF;
            }
        }

        /// <summary>
        /// Gets/sets size of this screen's item data.
        /// </summary>
        public byte ScreenEntrySize {
            get {
                return data[screenOffset + 1];
            }
            set {
                data[screenOffset + 1] = value;
            }
        }

        /// <summary>
        /// Seeks to the next screen, if there are more screens in this item data entry.
        /// </summary>
        /// <returns>True if the seeker is pointing at another screen data after the operation, or false if the seeker is beyond the end of screen data.</returns>
        public bool NextScreen() {
            bool hasMore = this.MoreScreens;

            screenOffset += ScreenEntrySize;
            itemOffset = screenOffset + 2;

            return hasMore;
        }
    }

    /// <summary>Represents the different types of items.</summary>
    public enum ItemType:byte{
        /// <summary>No item.</summary>
        Nothing = 0,
        /// <summary>Enemy.</summary>
        Enemy = 1,
        /// <summary>Power up</summary>
        PowerUp = 2,
        /// <summary>Unknown.</summary>
        Mella = 3,
        /// <summary>Elevator.</summary>
        Elevator = 4,
        /// <summary>Unknown.</summary>
        Unknown_5 = 5,
        /// <summary>Bridge to Tourain elevator.</summary>
        AccessBridge = 6,
        /// <summary>Enemy that respawns.</summary>
        RespawnEnemy = 7,
        /// <summary>Unknown.</summary>
        Rinkas = 8,
        /// <summary>Unknown.</summary>
        Door = 9,
        /// <summary>Palette swap.</summary>
        PalSwap = 0xA,
        /// <summary>Unknown.</summary>
        Unknown_b = 0xB,
        /// <summary>Unknown.</summary>
        Unknown_c = 0xC,
        /// <summary>Unknown.</summary>
        Unknown_d = 0xD,
        /// <summary>Unknown.</summary>
        Unknown_e = 0xE,
        /// <summary>Unknown.</summary>
        Unknown_f = 0xF,
    }

    /// <summary>Represents different power ups.</summary>
    public enum PowerUpType:byte
    {
        /// <summary>Maru mari bombs.</summary>
        Bomb = 0,
        /// <summary>High jump boots.</summary>
        HiJump = 1,
        /// <summary>Long beam.</summary>
        LongBeam = 2,
        /// <summary>Screw attack.</summary>
        ScrewAttack = 3,
        /// <summary>Maru Mari ("morph ball")</summary>
        MaruMari = 4,
        /// <summary>Varia suit.</summary>
        Varia = 5,
        /// <summary>Wave beam.</summary>
        WaveBeam = 6,
        /// <summary>Ice beam.</summary>
        IceBeam = 7,
        /// <summary>Energy tank.</summary>
        EnergyTank = 8,
        /// <summary>Missile expansion.</summary>
        Missile = 9,
        /// <summary>Missile expansion with incorrect graphics.</summary>
        Invalid_Missile_Dot = 0xA,
        /// <summary>Missile expansion with incorrect graphics.</summary>
        Invalid_Missile_Arrow = 0xB,
        /// <summary>Unknown.</summary>
        Unknown_c = 0xC,
        /// <summary>Unknown.</summary>
        Unknown_d = 0xD,
        /// <summary>Unknown.</summary>
        Unknown_e = 0xE,
        /// <summary>Unknown.</summary>
        Unknown_f = 0xF
    }

    /// <summary>Represents destinations for elevators.</summary>
    public enum ElevatorDestination
    {
        /// <summary>Brinstar.</summary>
        Brinstar = 1,
        /// <summary>Norfair.</summary>
        Norfair = 2,
        /// <summary>Kraid.</summary>
        Kraid = 3,
        /// <summary>Tourain.</summary>
        Tourain = 4,
        /// <summary>Ridley.</summary>
        Ridley = 5,
        /// <summary>GameEnd.</summary>
        GameEnd = 0xF,

        /// <summary>Flag. Indicates that elevator moves up instead of down.</summary>
        ElevatorUp = 0x80

    }

    /// <summary>
    /// Represents the screen coordinate of an item, object, or enemy.
    /// </summary>
    public struct ScreenCoordinate
    {
        byte val;
        /// <summary>
        /// The X coordinate this object represents.
        /// </summary>
        public int X {
            get {
                return val & 0xF; 
            }
            set {
                val = (byte)((val & 0xF0) | (value & 0xF)); 
            }
        }
        /// <summary>
        /// The Y coordinate this object represents.
        /// </summary>
        public int Y {
            get {
                return (val & 0xF0) / 0x10;
            }
            set {
                val = (byte)((val & 0x0F) | ((value * 0x10) & 0xFF));
            }
        }

        /// <summary>
        /// Instantiates this object.
        /// </summary>
        /// <param name="value">The byte that contains the data.</param>
        public ScreenCoordinate(byte value) {
            val = value;
        }
        /// <summary>
        /// Instantiates this object.
        /// </summary>
        /// <param name="x">X position.</param>
        /// <param name="y">Y position.</param>
        public ScreenCoordinate(int x, int y) {
            val = 0;
            X = x;
            Y = y;
        }
        /// <summary>
        /// Gets/sets the data for this object.
        /// </summary>
	    public byte Value
	    {
		    get { return val;}
		    set { val = value;}
	    }

    }
}
