using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Editroid.ROM;

namespace Editroid
{
	/// <summary>
	/// This class encapsulates level specific data.
	/// </summary>
	/// <remarks>This class stores some data locally. Certain changes will not necessarily be reflected in the ROM.</remarks>
	public class Level
	{
		byte[] _Data;
		LevelPointers _Pointers;
		Levels _Level;
		ComboTable _Combos;
		PatternTable _Patterns;
		PatternTable _SpritePatterns;
		Graphic.Sprite[] _Sprites;
        //ComboPhysics comboPhysics;

        /// <summary>Gets a Levels enumeration value for this level.</summary>
        public Levels LevelIdentifier {
            get { return _Level; }
        }

		NesPalette _BgPalette;
		/// <summary>Gets a NesPalette object representing the background 
		/// palette for this level.</summary>
		public NesPalette BgPalette { get { return _BgPalette; } }
		NesPalette _BgAltPalette;
		/// <summary>Gets a NesPalette object representing the alternate background 
		/// palette for this level.</summary>
		public NesPalette BgAltPalette { get { return _BgAltPalette; } }
		NesPalette _SpritePalette;
		/// <summary>Gets a NesPalette object representing the sprite
		/// palette for this level.</summary>
		public NesPalette SpritePalette { get { return _SpritePalette; } }
        /// <summary>
        /// Gets the combos for this level.
        /// </summary>
		public ComboTable Combos { get { return _Combos; } }

        MetroidPointer LevelStartPointer = new MetroidPointer(0xD7, 0x95);
        /// <summary>
        /// Gets the offset of the level start screen.
        /// </summary>
        public int LevelStartOffset { get { return LevelStartPointer.GetDataOffset(_Level); } }

        /// <summary>
        /// Gets/sets the location where the player will start in this level if he dies/continues.
        /// </summary>
        public int StartScreenX {
            get { return _Data[LevelStartOffset]; }
            set { _Data[LevelStartOffset] = (byte)value; }
        }
        /// <summary>
        /// Gets/sets the location where the player will start in this level if he dies/continues.
        /// </summary>
        public int StartScreenY {
            get { return _Data[LevelStartOffset + 1]; }
            set { _Data[LevelStartOffset + 1] = (byte)value; }
        }

		private void LoadRooms() {
			_Screens = new Screen[_Pointers.rooms.Length];

			// Get the location of the first screen pointer
			int PointerOffset = _Pointers.rooms.Offset;

			// For all screen pointers
			for(int i = 0; i < _Pointers.rooms.Length; i++) {
				// Load the pointer
				MetroidPointer ScreenPointer = new MetroidPointer(_Data, PointerOffset);

				// Resolve the pointer to a pointer and use that to load the room
				int ScreenPointerOffset = ScreenPointer.GetDataOffset(_Level);
                //_Screens[i] = new Screen(_Data, ScreenPointer);
                _Screens[i] = new Screen(_Data, new MetroidHandle(_Data, PointerOffset, this._Level));
                PointerOffset += 2;
            }
		}

		private void LoadStructs() {
			_Structs = new Struct[_Pointers.structs.Length];

			// Get the location of the first struct pointer
			int PointerOffset = _Pointers.structs.Offset;

			// For all struct pointers
			for(int i = 0; i < _Pointers.structs.Length; i++) {
				// Load the pointer
				MetroidPointer StructPointer = new MetroidPointer(_Data, PointerOffset);
				PointerOffset += 2;

				// Resolve the pointer to an offset and use that to load the struct
				int StructPointerOffset = StructPointer.GetDataOffset(_Level);
				_Structs[i] = new Struct(_Data, StructPointerOffset);
			}
		}

        /// <summary>
        /// Gets the offset of structure data.
        /// </summary>
        public int StructDataOffset {
            get {
                return _Structs[0].Offset;
            }
        }
        /// <summary>
        /// Removes data from a screen entry and compacts screen data. No safety checks are performed.
        /// </summary>
        /// <param name="screen">The screen to remove data from</param>
        /// <param name="DataOffset">The offset of the data to remove.</param>
        /// <param name="DataSize">The number of bytes to remove.</param>
        public void CropScreenData(Screen screen, int DataOffset, int DataSize) {
            // Determine screen index
            int screenIndex = 0;
            while(screenIndex < ScreenCount && _Screens[screenIndex].Offset != screen.Offset) {
                screenIndex++;
            }

            if(screenIndex == ScreenCount)                 throw new ArgumentException("Specified screen is not valid, could not be located in level data.");

            // Copy screen data from DataOffset to END of level screen data back 'DataSize' bytes.
            for(int i = DataOffset; i < StructDataOffset - DataSize; i++) {
                _Data[i] = _Data[i + DataSize];
            }

            // Move all handles following 'screen's back 'DataSize' bytes.
            for(int i = screenIndex + 1; i < ScreenCount; i++) {
                _Screens[i].ScreenHandle.OffsetPointer(-DataSize);
            }

            LoadRooms();
        }
        
        /// <summary>
        /// Expands the screen data black and inserts data into a screen entry. No safety checks are performed. New memory is not initialized.
        /// </summary>
        /// <param name="screen">The screen to insert data into.</param>
        /// <param name="DataOffset">The offset to add data at.</param>
        /// <param name="DataSize">The number of bytes to add.</param>
        public void ExpandScreenData(Screen screen, int DataOffset, int DataSize) {
            // Determine screen index
            int screenIndex = 0;
            while(screenIndex < ScreenCount && _Screens[screenIndex].Offset != screen.Offset) {
                screenIndex++;
            }

            if(screenIndex == ScreenCount) throw new ArgumentException("Specified screen is not valid, could not be located in level data.");

            // Copy screen data from END of level screen data to DataOffset ahead 'DataSize' bytes.
            for(int i = StructDataOffset - 1; i >= DataOffset + DataSize; i--) {
                _Data[i] = _Data[i - DataSize];
            }

            // Move all handles following 'screen's back 'DataSize' bytes.
            for(int i = screenIndex + 1; i < ScreenCount; i++) {
                _Screens[i].ScreenHandle.OffsetPointer(DataSize);
            }

            LoadRooms();
        }
        /// <summary>
        /// Gets the remaining free space for room data for this level, assuming there is no data between
        /// room data and structure data.
        /// </summary>
        public int FreeRoomData {
            get {
                Screen lastRoom = GetScreen(ScreenCount - 1);
                // 3 bytes per object/enemy, 2 bytes per door, 3 bytes for general room data
                int size = lastRoom.Objects.Length * 3 + lastRoom.Enemies.Length * 3 + lastRoom.Doors.Length * 2 + 3;
                return GetStruct(0).Offset - lastRoom.Offset - size;
            }
        }

		/// <summary>
		/// Loads all level-specific data for a specific level in a given ROM image
		/// </summary>
		/// <param name="rom">The Rom object to load data from</param>
		/// <param name="level">The level to load</param>
		public Level(Rom rom, Levels level) {
			_Data = rom.data;
			_Pointers = LevelPointers.GetPointers(level);
			_Level = level;

			LoadStructs();
			LoadRooms();
			LoadPatterns();
			LoadPalette();
			_Combos = new ComboTable(_Data, Combo.GetComboOffset(_Level));
			_Sprites = Graphic.Sprite.GetSprites(_Level);
            itemTable = new ItemTable(_Level, _Data);
            //comboPhysics = ComboPhysics.GetPhysics(level);
		}

        private ItemTable itemTable;

        /// <summary>
        /// Gets the item table for this level.
        /// </summary>
        public ItemTable ItemTable {
            get { return itemTable; }
        }


        //public ComboPhysics ComboPhysicsTable { get { return comboPhysics; } }

		/// <summary>
		/// Gets the number of sprites in this level.
		/// </summary>
		public int SpriteCount { get { return _Sprites.Length; } }
		/// <summary>
		/// Gets a sprite by index.
		/// </summary>
		/// <param name="index">Index of the sprite to get.</param>
		/// <returns>A Graphic.Sprite object.</returns>
		public Graphic.Sprite GetSprite(int index) {
			if(index < _Sprites.Length)
				return _Sprites[index];
			//Todo: throw exception on invalid sprite
			return _Sprites[0];
		}

		private void LoadPalette() {
			_BgPalette = new NesPalette(_Data, NesPalette.GetBackgroundOffset(_Level));
			_BgAltPalette = new NesPalette(_Data, NesPalette.GetAltBackgroundOffset(_Level));
			_SpritePalette = new NesPalette(_Data, NesPalette.GetSpriteOffset(_Level));
		}

		/// <summary>
		/// Gets the pattern table for this level.
		/// </summary>
		public PatternTable Patterns { get { return _Patterns; } }
		/// <summary>
		/// Gets the sprite pattern table for this level.
		/// </summary>
		public PatternTable SpritePatterns { get { return _SpritePatterns; } }

        /// <summary>
        /// Gets a linear pattern table that can be used for blitting to construct images using level-specific graphics. The table is 1 graphic tile tall and 256 tiles wide.
        /// </summary>
		private void LoadPatterns() {
			_Patterns = new PatternTable(true);
			PatternTableRange[] tileSource = PatternTableRange.GetBackgroundTableRanges(_Level);

			_Patterns.BeginWrite();

			//int TileOffset = 0;
			//for(int i = 0; i < tileSource.Length; i++) {
			//    _Patterns.LoadTiles(_Data, tileSource[i].RomOffset, TileOffset, tileSource[i].TileCount);
			//    TileOffset += tileSource[i].TileCount;
			//}
			_Patterns.LoadTiles(_Data, tileSource);

			_Patterns.EndWrite();

			_SpritePatterns = new PatternTable(true);
			tileSource = PatternTableRange.GetSpriteTableRange(_Level);

			_SpritePatterns.BeginWrite();
			//int TileOffset = 0;
			//for(int i = 0; i < tileSource.Length; i++) {
			//    _SpritePatterns.LoadTiles(_Data, tileSource[i].RomOffset, TileOffset, tileSource[i].TileCount);
			//    TileOffset += tileSource[i].TileCount;
			//}
			_SpritePatterns.LoadTiles(_Data, tileSource);

			_SpritePatterns.EndWrite();

		}


		Struct[] _Structs;
		Screen[] _Screens;

		/// <summary>
		/// Gets a struct definition.
		/// </summary>
		/// <param name="index">Index of a struct.</param>
		/// <returns>A struct definition.</returns>
		public Struct GetStruct(int index) {
			// TOdo: validate index
			if(index > _Structs.Length) return _Structs[0];
			return _Structs[index];
		}

		/// <summary>
		/// Gets a combo in this level by index.
		/// </summary>
		/// <param name="index">Index of the combo.</param>
		/// <returns>A Combo object.</returns>
		public Combo GetCombo(int index) {
			return _Combos[index];
		}


		/// <summary>
		/// Gets the number of structs in this level.
		/// </summary>
		public int StructCount {
			get { return _Structs.Length; }
		}
		/// <summary>
		/// Gets the number of screens in this level.
		/// </summary>
		public int ScreenCount {
			get { return _Screens.Length; }
		}
		/// <summary>
		/// Gets screen specific data.
		/// </summary>
		/// <param name="index">The index of a screen.</param>
		/// <returns>Screen specific data</returns>
		public Screen GetScreen(int index) {
			// TOdo: validate index
			if(index >= _Screens.Length) return (_Screens[1]);
			return _Screens[index];
		}

        internal void ReloadItems() {
            itemTable = new ItemTable(_Level, _Data);
        }
    }
}
