using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace Editroid
{
	/// <summary>
	/// Represents a metroid pointer.
	/// </summary>
    /// <remarks>Metroid pointers are often relative, particularly level-specific data.
    /// Use the GetDataOffset method to obtain the correctly offset pointer value.</remarks>
	public struct Mointer
	{
		internal Byte Byte1, Byte2;

		/// <summary>
		/// Creates a mointer from a composite value
		/// </summary>
		/// <param name="value">An unsigned 16-bit composite of two bytes making up a mointer</param>
		public Mointer(ushort value) {
			byte[] bytes = BitConverter.GetBytes(value);
			Byte1 = bytes[1];
			Byte2 = bytes[0];
		}
		/// <summary>
		/// Creates a mointer from a composite value
		/// </summary>
		/// <param name="value">A 16-bit composite of two bytes making up a mointer</param>
		public Mointer(short value) {
			byte[] bytes = BitConverter.GetBytes(value);
			Byte1 = bytes[1];
			Byte2 = bytes[0];
		}
		/// <summary>
		/// Creates a mointer from two bytes
		/// </summary>
		/// <param name="byte1">First byte of this Mointer</param>
		/// <param name="byte2">Second byte of this Mointer</param>
		public Mointer(byte byte1, byte byte2) {
			Byte1 = byte1;
			Byte2 = byte2;
		}
		/// <summary>
		/// Loads a mointer from raw data
		/// </summary>
		/// <param name="source">Source of the data</param>
		/// <param name="offset">The offset of the mointer</param>
		public Mointer(byte[] source, int offset) {
			Byte1 = source[offset];
			Byte2 = source[offset + 1];
		}
		/// <summary>
		/// Gets the actual value represented by this mointer.
		/// </summary>
		/// <remarks>This is not necessarily the offset of the 
		/// data that the mointer is intended to refer to. If the data is
		/// level-specific, use the GetDataOffset function. This corrects
		/// offsets not reflected in the raw mointer.</remarks>
		public int RealValue {
			get {
				return Byte1 | ((int)Byte2 << 8);
			}
			set {
				Byte2 = (byte)((value >> 8) & 0xFF);
				Byte1 = (byte)(value & 0xFF);
			}
		}

		/// <summary>
		/// Gets the offset of the level-specific data that this mointer
		/// refers to.
		/// </summary>
		/// <param name="level">The level this mointer refers to</param>
		/// <returns>the offset of the level-specific data that this mointer
		/// refers to</returns>
		/// <remarks>This function only applies to certain level-specific data.</remarks>
		public int GetDataOffset(Levels level) {
            if(level == Levels.Norfair || level == Levels.Tourain)
                return RealValue + GetLevelOffset(level) + 0x8000;
            else if(level == Levels.Ridley)
                return RealValue + GetLevelOffset(level) + 0xC000;
            else if(level == Levels.Kraid)
                return RealValue + GetLevelOffset(level) + 0x4000;
            else if(level == Levels.Brinstar)
                return RealValue + GetLevelOffset(level);
            else
                return RealValue;
		}
		/// <summary>
		/// Calculates and stores the correct mointer value to point to 
		/// level-specific data
		/// </summary>
		/// <param name="level">The level this mointer refers to</param>
		/// <param name="offset">The offset of the actual data in the ROM</param>
		public void SetDataOffset(Levels level, int offset) {
			RealValue = offset - GetLevelOffset(level);
			//Check: Test SetDataOffset() (there may be issues with 4-digit pointer wrapping);
		}

		/// <summary>
		/// Mointer offset correction for Brinstar. This value does not account for different memory banks.
		/// </summary>
		public static int BrinstarOffset { get { return -16368; } }
		/// <summary>
		/// Mointer offset correction for Ridley's Hideout. This value does not account for different memory banks.
		/// </summary>
		public static int RidleyOffset { get { return 16; } }
		/// <summary>
		/// Mointer offset correction for Kraid's Hideout. This value does not account for different memory banks.
		/// </summary>
		public static int KraidOffset { get { return 16400; } }
		/// <summary>
		/// Mointer offset correction for Norfair. This value does not account for different memory banks.
		/// </summary>
		public static int NorfairOffset { get { return -32752; } }
		/// <summary>
		/// Mointer offset correction for Tourain. This value does not account for different memory banks.
		/// </summary>
		public static int TourainOffset { get { return -16368; } }
		/// <summary>
		/// Gets the mointer offset correction for a specified level.
		/// This value does not account for different memory banks.
		/// </summary>
		/// <param name="level">The level to get the correction value for</param>
		/// <returns>The mointer offset correction</returns>
		public static int GetLevelOffset(Levels level) {
			switch(level) {
				case Levels.Brinstar:
					return BrinstarOffset;
				case Levels.Kraid:
					return KraidOffset;
				case Levels.Norfair:
					return NorfairOffset;
				case Levels.Ridley:
					return RidleyOffset;
				case Levels.Tourain:
					return TourainOffset;
				case Levels.Invalid:
					return 0;
			}
			throw new ArgumentException("Invalid level");
		}
	}


    /// <summary>
    /// "Metroid handle." Handle points to a ROM offset that contains a pointer to ROM data, allowing access to the data and modification of the pointer.
    /// </summary>
    public struct Mandle
    {
        //Rom rom;
        byte[] data;
        int offset;
        Levels level;

        /// <summary>
        /// Gets/sets which level this handle references data for.
        /// </summary>
        public Levels Level {
            get { return level; }
            set { level = value; }
        }

        /// <summary>Gets the location in the rom where the actual pointer is stored.</summary>
        public int HandleOffset { get { return offset; } }
        /// <summary>Gets the uncorrected value of thie pointer this handle references.</summary>
        public int PointerValue {
            get {
                return Pointer.GetDataOffset(Levels.Invalid);
            }
        }
        /// <summary>Gets the ROM offset for the data that this handle's pointer references.</summary>
        public int PointerTarget {
            get {
                return Pointer.GetDataOffset(level);
            }
        }

        /// <summary>Instantiates this object.</summary>
        /// <param name="offset">Location of pointer in ROM data.</param>
        /// <param name="data">Rom image this pointer points to.</param>
        public Mandle(byte[] data, int offset) {
            this.data = data;
            this.offset = offset;
            level = Levels.Invalid;
        }

        /// <summary>Instantiates this object.</summary>
        /// <param name="data">ROM data.</param>
        /// <param name="offset">Location of pointer in ROM data.</param>
        /// <param name="level">The level that the pointer refers to data for.</param>
        public Mandle(byte[] data, int offset, Levels level) {
            this.data = data;
            this.offset = offset;
            this.level = level;
        }

        /// <summary>
        /// Gets the metroid pointer that this handle references.
        /// </summary>
        public Mointer Pointer {
            get {
                return new Mointer(data[offset], data[offset + 1]);
            }
        }

        /// <summary>
        /// Modifies the underlying pointer, allowing pointer arithmetic to be performed.
        /// </summary>
        /// <param name="amount">The amount to offset the pointer.</param>
        public void OffsetPointer(int amount) {
            // Get pointer data
            Union_16 pointerValue = Union_16.Create();
            pointerValue.Byte2 = data[offset + 1];
            pointerValue.Byte1 = data[offset];
            // Offset pointer data
            pointerValue.UShort = (ushort)(pointerValue.UShort + amount);
            // Store modified pointer data
            data[offset + 1] = pointerValue.Byte2;
            data[offset] = pointerValue.Byte1;
        }
    }

    /// <summary>
    /// Allows direct access to the same memory as different data types.
    /// </summary>
    /// <remarks>Editing one member of this struct will modify the value of some
    /// or all other members of this struct. All members share the same memory.</remarks>
    [StructLayout(LayoutKind.Explicit)]
    public struct Union_16
    {
        /// <summary>The first byte of the 16 bits.</summary>
        [FieldOffset(0)] public Byte Byte1;
        /// <summary>The second byte of the 16 bits.</summary>
        [FieldOffset(1)] public Byte Byte2;
        /// <summary>The value of the 16 bits as an unsigned integer.</summary>
        [FieldOffset(0)] public ushort UShort;
        /// <summary>The value of the 16 bits as a signed integer.</summary>
        [FieldOffset(0)] public short Short;
        /// <summary>The value of the 16 bits as a unicode character.</summary>
        [FieldOffset(0)] public Char Char;

        /// <summary>Gets an initialized instance of this struct.</summary>
        /// <returns>An initialized instance of this struct.</returns>
        public static Union_16 Create() {
            Union_16 result;

            result.UShort = 0;
            result.Short = 0;
            result.Char = '\0';
            result.Byte1 = result.Byte2 = 0;

            return result;
        }
    }
}
