using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Editroid
{
	/// <summary>
	/// Provides functions to load ROM graphic data into a System.Drawing.Bitmap.
	/// </summary>
	public class PatternTable
	{
		Bitmap patterns;
		bool _Linear;

		/// <summary>
		/// Creates a new pattern table.
		/// </summary>
		/// <remarks>This overload creates a rectangular pattern table</remarks>
		public PatternTable() 
		:this(false) { }

		/// <summary>
		/// Creates a new pattern table.
		/// </summary>
		/// <param name="linear">If true, the patterns will be loaded in a horizontal 
		/// string (good for constructing images from). If false, the patterns will be
		/// loaded in a 16 x 16 grid (good for displaying to user).</param>
		public PatternTable(bool linear) {
			_Linear = linear;
			if(linear)
				patterns = new Bitmap(2048, 8, PixelFormat.Format8bppIndexed);
			else
				patterns = new Bitmap(128, 128, PixelFormat.Format8bppIndexed);
		}

		/// <summary>
		/// Creates a new pattern table.
		/// </summary>
		/// <param name="linear">If true, the patterns will be loaded in a horizontal 
		/// string (good for constructing images from). If false, the patterns will be
		/// loaded in a 16 x 16 grid (good for displaying to user).</param>
        /// <param name="data">ROM data</param>
        /// <param name="tileSources">Locations of tiles</param>
		public PatternTable(bool linear, byte[] data, PatternTableRange[] tileSources) {
			_Linear = linear;
			if(linear)
				patterns = new Bitmap(2048, 8, PixelFormat.Format8bppIndexed);
			else
				patterns = new Bitmap(128, 128, PixelFormat.Format8bppIndexed);

			this.BeginWrite();
			this.LoadTiles(data, tileSources);
			this.EndWrite();
		}
		BitmapData lockData = null;
		byte[] lockBits = null;

		/// <summary>
		/// Prepares the pattern table for loading tiles.
		/// </summary>
		public void BeginWrite() {
			if(lockData != null) throw new InvalidOperationException("The pattern table is already opened for writing.");

			// Opened as Read/Write so that old, unmodified data is left intact
			lockData = patterns.LockBits(new Rectangle(0, 0, patterns.Width, patterns.Height), ImageLockMode.ReadWrite, PixelFormat.Format8bppIndexed);
			if(lockBits == null) lockBits = new byte[0x4000];

			Marshal.Copy(lockData.Scan0, lockBits, 0, 0x4000);
		}

		/// <summary>
		/// Finalizing writing operations to the pattern table.
		/// </summary>
		public void EndWrite() {
			if(lockData == null) throw new InvalidOperationException("The pattern table is not opened for writing.");

			Marshal.Copy(lockBits, 0, lockData.Scan0, 0x4000);
			patterns.UnlockBits(lockData);
		}

		/// <summary>
		/// Loads a strip of tiles from raw ROM data into this pattern table.
		/// </summary>
		/// <param name="data">The ROM image to load tiles from</param>
		/// <param name="offset">The location of the graphic data</param>
		/// <param name="tileOffset">Which tile to begin loading at</param>
		/// <param name="tileCount">The number of tiles to load</param>
		/// <remarks>A linear strip of tiles is loaded starting at the specified tile index. If the tiles
		/// reach the end of a row in a square pattern table they wrap to the next line. If the end of the pattern table is reached
		/// or the end of ROM data is reached, copying will stop without error.
		/// The function BeginWrite must be called at some point before this function
		/// is called. The function EndWrite must be called to finalize the process. Otherwise
		/// the loaded tiles will not be reflected in this pattern table.</remarks>
		public void LoadTiles(byte[] data, int offset, int tileOffset, int tileCount) {
			if(_Linear)
				LoadTilesString(data, offset, tileOffset, tileCount);
			else
				LoadTilesGrid(data, offset, tileOffset, tileCount);
		}

		/// <summary>
		/// Loads a strip of tiles from raw ROM data into this pattern table.
		/// </summary>
		/// <param name="data">The ROM image to load tiles from</param>
		/// <param name="tileOffset">Where to load the tiles into the pattern table</param>
		/// <param name="tiles">A PatternTableRange that specifies a region of data
		/// to load from the ROM</param>
		/// <remarks>The function BeginWrite must be called at some point before this function
		/// is called. The function EndWrite must be called to finalize the process. Otherwise
		/// the loaded tiles will not be reflected in this pattern table.</remarks>
		public void LoadTiles(byte[] data, int tileOffset, PatternTableRange tiles) {
			if(_Linear)
				LoadTilesString(data, tiles.RomOffset, tileOffset, tiles.TileCount);
			else
				LoadTilesGrid(data, tiles.RomOffset, tileOffset, tiles.TileCount);
		}

		/// <summary>
		/// Uses an array of PatternTableRange objects to fill a pattern table.
		/// </summary>
		/// <param name="data">ROM image to load tiles from</param>
		/// <param name="source">An array of PatternTableRange objects that provides
		/// the source to load data from</param>
		/// <remarks>The function BeginWrite must be called at some point before this function
		/// is called. The function EndWrite must be called to finalize the process. Otherwise
		/// the loaded tiles will not be reflected in this pattern table.</remarks>
		public void LoadTiles(byte[] data, PatternTableRange[] source) {
			int TileOffset = 0;
			for(int i = 0; i < source.Length; i++) {
				int TileCount = source[i].TileCount;
				if(TileOffset + TileCount > 256) TileCount = 256 - TileOffset;
				LoadTiles(data, source[i].RomOffset, TileOffset, TileCount);
				TileOffset += TileCount;
			}
		}

		private void LoadTilesGrid(byte[] data, int offset, int tileOffset, int tileCount) {
			//Todo: LoadTiles safety checks
			for(int Tile = tileOffset; Tile < tileOffset + tileCount; Tile++) {
				// The tile X/Y coords are used to find the pixel offset within the resulting image
				int TileY = Tile / 16;
				int TileX = Tile % 16;

				// NES graphic data is bi-planar (2bpp image is actually stored as two 1bpp images)
				// Load first plane
				for(int PixelY = 0; PixelY < 8; PixelY += 1) {
					int imageOffset = TileY * 1024 + TileX * 8 + PixelY * 128;
					lockBits[imageOffset + 7] = (byte)(data[offset] & 0x01);
					lockBits[imageOffset + 6] = (byte)((data[offset] & 0x02) >> 1);
					lockBits[imageOffset + 5] = (byte)((data[offset] & 0x04) >> 2);
					lockBits[imageOffset + 4] = (byte)((data[offset] & 0x08) >> 3);
					lockBits[imageOffset + 3] = (byte)((data[offset] & 0x10) >> 4);
					lockBits[imageOffset + 2] = (byte)((data[offset] & 0x20) >> 5);
					lockBits[imageOffset + 1] = (byte)((data[offset] & 0x40) >> 6);
					lockBits[imageOffset + 0] = (byte)((data[offset] & 0x80) >> 7);
					offset++;

				}
				for(int PixelY = 0; PixelY < 8; PixelY += 1) {
					int imageOffset = TileY * 1024 + TileX * 8 + PixelY * 128;
					lockBits[imageOffset + 7] |= (byte)((data[offset] & 0x01) << 1);
					lockBits[imageOffset + 6] |= (byte)((data[offset] & 0x02));
					lockBits[imageOffset + 5] |= (byte)((data[offset] & 0x04) >> 1);
					lockBits[imageOffset + 4] |= (byte)((data[offset] & 0x08) >> 2);
					lockBits[imageOffset + 3] |= (byte)((data[offset] & 0x10) >> 3);
					lockBits[imageOffset + 2] |= (byte)((data[offset] & 0x20) >> 4);
					lockBits[imageOffset + 1] |= (byte)((data[offset] & 0x40) >> 5);
					lockBits[imageOffset + 0] |= (byte)((data[offset] & 0x80) >> 6);

					offset++;
				} // for
			} // for
		} // function
		private void LoadTilesString(byte[] data, int offset, int tileOffset, int tileCount) {
			//Todo: LoadTiles safety checks
			for(int Tile = tileOffset; Tile < tileOffset + tileCount; Tile++) {
				// NES graphic data is bi-planar (2bpp image is actually stored as two 1bpp images)
				// Load first plane
				for(int PixelY = 0; PixelY < 8; PixelY += 1) {
					int imageOffset = Tile * 8 + PixelY * 2048;
					lockBits[imageOffset + 7] = (byte)(data[offset] & 0x01);
					lockBits[imageOffset + 6] = (byte)((data[offset] & 0x02) >> 1);
					lockBits[imageOffset + 5] = (byte)((data[offset] & 0x04) >> 2);
					lockBits[imageOffset + 4] = (byte)((data[offset] & 0x08) >> 3);
					lockBits[imageOffset + 3] = (byte)((data[offset] & 0x10) >> 4);
					lockBits[imageOffset + 2] = (byte)((data[offset] & 0x20) >> 5);
					lockBits[imageOffset + 1] = (byte)((data[offset] & 0x40) >> 6);
					lockBits[imageOffset + 0] = (byte)((data[offset] & 0x80) >> 7);
					offset++;

				}
				for(int PixelY = 0; PixelY < 8; PixelY += 1) {
					int imageOffset = Tile * 8 + PixelY * 2048;
					lockBits[imageOffset + 7] |= (byte)((data[offset] & 0x01) << 1);
					lockBits[imageOffset + 6] |= (byte)((data[offset] & 0x02));
					lockBits[imageOffset + 5] |= (byte)((data[offset] & 0x04) >> 1);
					lockBits[imageOffset + 4] |= (byte)((data[offset] & 0x08) >> 2);
					lockBits[imageOffset + 3] |= (byte)((data[offset] & 0x10) >> 3);
					lockBits[imageOffset + 2] |= (byte)((data[offset] & 0x20) >> 4);
					lockBits[imageOffset + 1] |= (byte)((data[offset] & 0x40) >> 5);
					lockBits[imageOffset + 0] |= (byte)((data[offset] & 0x80) >> 6);

					offset++;
				} // for
			} // for
		} // function

		/// <summary>
		/// Returns an 8 bit-per-pixel bitmap containing all loaded tiles.
		/// </summary>
		public Bitmap Patterns {
			get {
				return patterns;
			}
		}

		/// <summary>
		/// Gets or sets the palette applied to the tiles.
		/// </summary>
		/// <remarks>Changes to the palette will be reflected in all references to
		/// this PatternTable's Patterns. If you need to use multiple palettes simultaneously
		/// you should, for each differently paletted version, set the desired palette,
		/// clone the pattern bitmap and retain the clone(s) as the paletted bitmap(s).</remarks>
		public Color[] Palette {
			get {
				return patterns.Palette.Entries;
			}
			set {
				ColorPalette pal = patterns.Palette;
				if(value.Length != 4) throw new InvalidOperationException("Only a four-color palette can be specified.");

				Array.Copy(value, pal.Entries, 4);
				patterns.Palette = pal;
			}
		}
	}

	/// <summary>
	/// Defines a range of ROM memory to be loaded into a Pattern Table.
	/// </summary>
	/// <remarks>This class is primarily intended to encapsulate the pattern table
	/// data ranges for Metroid using the public fields "Brinstar", "Norfiar", etc., but
	/// it can be used to specify data ranges and load tiles for any NES graphic data.</remarks>
	public struct PatternTableRange
	{
		int _RomOffset, _TileCount;
		/// <summary>
		/// Instantiates this object.
		/// </summary>
		/// <param name="romOffset">The offset of the graphic data within the ROM.</param>
		/// <param name="tileCount">The number of tiles included in this range.</param>
		public PatternTableRange(int romOffset, int tileCount) {
			_RomOffset = romOffset;
			_TileCount = tileCount;
		}

		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] Brinstar = new PatternTableRange[] {
			new PatternTableRange(0x19Db0, 0x15),  // 15
			new PatternTableRange(0x0, 0xB),       // 20
			new PatternTableRange(0x14D70, 0x45),  // 65
			new PatternTableRange(0x0, 0x1B),      // 80
			new PatternTableRange(0x19F00, 0x80)
		};

        /// <summary>
        /// Gets the pattern table offsets for the specified level.
        /// </summary>
        /// <param name="level">A level.</param>
        /// <returns>An array of PatternTableRange objects.</returns>
        public static PatternTableRange[] GetRanges(Levels level) {
            switch(level) {
                case Levels.Brinstar:
                    return Brinstar;
                case Levels.Norfair:
                    return Norfair;
                case Levels.Tourain:
                    return Tourain;
                case Levels.Kraid:
                    return Kraid;
                case Levels.Ridley:
                    return Ridley;
                default:
                    throw new ArgumentException("The specified level was invalid.", "level");
            }
        }
        /// <summary>
        /// Gets the pattern table offsets for the specified level.
        /// </summary>
        /// <param name="level">A level.</param>
        /// <returns>An array of PatternTableRange objects.</returns>
        public static PatternTableRange[] GetSpriteRanges(Levels level) {
            switch(level) {
                case Levels.Brinstar:
                    return BrinstarSprites;
                case Levels.Norfair:
                    return NorfairSprites;
                case Levels.Tourain:
                    return TourainSprites;
                case Levels.Kraid:
                    return KraidSprites;
                case Levels.Ridley:
                    return RidleySprites;
                default:
                    throw new ArgumentException("The specified level was invalid.", "level");
            }
        }
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] BrinstarSprites = new PatternTableRange[] {
			new PatternTableRange(0x18010, 0x7B),
			new PatternTableRange(0x187C0, 0x1F),
			new PatternTableRange(0x00000, 0x06),
			new PatternTableRange(0x1B4D0, 0x0A),
			new PatternTableRange(0x00000, 0x16),
			new PatternTableRange(0x05170, 0x40)
		};

		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] Norfair = new PatternTableRange[] {
			new PatternTableRange(0x1A700, 0x26), // 26
			new PatternTableRange(0x14DD0, 0x3F), // 65
			new PatternTableRange(0x00, 0xB),     // 70
			new PatternTableRange(0x1A960, 0x7),  // 77
			new PatternTableRange(0x0, 0x9),      // 80
			new PatternTableRange(0x19F00, 0x80)  // 100
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] NorfairSprites = new PatternTableRange[] {
			new PatternTableRange(0x18010, 0x7B),
			new PatternTableRange(0x187C0, 0x1F),
			new PatternTableRange(0x00000, 0x06),
			new PatternTableRange(0x1B4D0, 0x0A),
			new PatternTableRange(0x00000, 0x16),
			new PatternTableRange(0x08D70, 0x40)
		};

		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] Tourain= new PatternTableRange[] {
			new PatternTableRange(0x1A9D0, 0x20), // 20
			new PatternTableRange(0x1ACB0, 0x60), // 80
			new PatternTableRange(0x19F00, 0x10), // 90
			new PatternTableRange(0x1B2B0, 0x9),  // 99
			new PatternTableRange(0x1A090, 0x37), // D0
			new PatternTableRange(0x1B4D0, 0x2A), // FA
			new PatternTableRange(0x1A6A0, 0x6)   // 100
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] TourainSprites = new PatternTableRange[] {
			new PatternTableRange(0x18010, 0x7B),
			new PatternTableRange(0x187C0, 0x1F),
			new PatternTableRange(0x00000, 0x06),
			new PatternTableRange(0x1B4D0, 0x0A),
			new PatternTableRange(0x00000, 0x16),
			new PatternTableRange(0x09170, 0x40)
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] Kraid = new PatternTableRange[] {
			new PatternTableRange(0x1A9D0, 0x30), // 30
			new PatternTableRange(0x14E70, 0x35), // 65
			new PatternTableRange(0x0, 0xB),      // 70
			new PatternTableRange(0x1B340, 0xC),  // 7C
			new PatternTableRange(0x0, 0x4),      // 80
			new PatternTableRange(0x19F00, 0x60), // E0
			new PatternTableRange(0x1B340, 0xC),  // EC
			new PatternTableRange(0x1A5C0, 0x14)  // 100
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] KraidSprites = new PatternTableRange[] {
			new PatternTableRange(0x18010, 0x7B),
			new PatternTableRange(0x187C0, 0x1F),
			new PatternTableRange(0x00000, 0x06),
			new PatternTableRange(0x1B4D0, 0x0A),
			new PatternTableRange(0x00000, 0x16),
			new PatternTableRange(0x0CD70, 0x40)
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] Ridley = new PatternTableRange[] {
			new PatternTableRange(0x1A9D0, 0x30), // 30
			new PatternTableRange(0x14E70, 0x35), // 65
			new PatternTableRange(0x0, 0xB),      // 70
			new PatternTableRange(0x1B400, 0xC),  // 7C
			new PatternTableRange(0x0, 0x4),      // 80
			new PatternTableRange(0x19F00, 0x60), // E0
			new PatternTableRange(0x1B400, 0xC),  // EC
			new PatternTableRange(0x1A5C0, 0x14)  // 100
		};
		/// <summary>
		/// Gets the ROM offsets and tile counts that should be used to construct a 
		/// 256-tile pattern table for this level.
		/// </summary>
		public static readonly PatternTableRange[] RidleySprites = new PatternTableRange[] {
			new PatternTableRange(0x18010, 0x7B),
			new PatternTableRange(0x187C0, 0x1F),
			new PatternTableRange(0x00000, 0x06),
			new PatternTableRange(0x1B4D0, 0x0A),
			new PatternTableRange(0x00000, 0x16),
			new PatternTableRange(0x0D170, 0x40)
		};

		/// <summary>
		/// Gets the PatternTableRanges that represent a complete pattern table
		/// for the specified level.
		/// </summary>
		/// <param name="level">a Metroid level</param>
		/// <returns>A PatternTableRanges that represent a complete pattern table.</returns>
		public static PatternTableRange[] GetBackgroundTableRanges(Levels level) {
			switch(level) {
				case Levels.Kraid:
					return Kraid;
				case Levels.Tourain:
					return Tourain;
				case Levels.Ridley:
					return Ridley;
				case Levels.Brinstar:
					return Brinstar;
				case Levels.Norfair:
					return Norfair;
				default:
					throw new ArgumentException("Specified level is invalid.", "level");
			}
		}
		/// <summary>
		/// Gets the PatternTableRanges for a specific level.
		/// </summary>
		/// <param name="level"></param>
		/// <returns></returns>
		public static PatternTableRange[] GetSpriteTableRange(Levels level) {
			switch(level) {
				case Levels.Kraid:
					return KraidSprites;
				case Levels.Tourain:
					return TourainSprites;
				case Levels.Ridley:
					return RidleySprites;
				case Levels.Brinstar:
					return BrinstarSprites;
				case Levels.Norfair:
					return NorfairSprites;
				default:
					throw new ArgumentException("Specified level is invalid.", "level");
			}
		}
		/// <summary>
		/// Gets the offset where graphic data begins.
		/// </summary>
		public int RomOffset { get { return _RomOffset; } }
		/// <summary>
		/// Gets the number of tiles included in the data range.
		/// </summary>
		public int TileCount { get { return _TileCount; } }
	}
}
