using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace Editroid
{
	/// <summary>
	/// Provides access to screen-specific data
	/// </summary>
	public class Screen
	{
		byte[] _Data;
		//int _Offset;
        int _Offset{
            get{
                return handle.PointerTarget;
            }
        }
        MetroidHandle handle;

        /// <summary>
        /// Gets the offset of enemy data (including door data). If there is
        /// no enemy or door data, the return value will be the offset of the
        /// first byte after this screen's data.
        /// </summary>
        public int EnemyDataOffset{
        get{
            return _Offset + // Start
                1 + // Default palette specifier
                Objects.Length * 3 +// Enemies
                1; // End of enemy data (FD) or end of data (FF) marker
            }
        }

        /// <summary>
        /// Gets a boolean value indicating whether the screen has a data
        /// segment for enemy data (identified by the hexadecimal value FD).
        /// </summary>
        public bool HasEnemyData {
            get {
                return _Data[EnemyDataOffset - 1] == 0xFD;
            }
        }

        /// <summary>
        /// Creates a Screen object.
        /// </summary>
        /// <param name="data">ROM data</param>
        /// <param name="handle">A metroid handle pointing to the data.</param>
        public Screen(Byte[] data, MetroidHandle handle) {
            _Data = data;
            this.handle = handle;
        }
        /// <summary>
        /// Gets the offset of this object in ROM data.
        /// </summary>
        public int Offset { get { return _Offset; } }

		/// <summary>
		/// Gets the ColorAttributeTable entry for this screen
		/// </summary>
		public byte ColorAttributeTable {
			get {
				return _Data[_Offset];
			}
			set {
				_Data[_Offset] = value;
			}
		}


        /// <summary>Gets a handle that refers to this screen's
        /// entry in the level's pointer table.</summary>
        public MetroidHandle ScreenHandle { get { return handle; } }

		ScreenObject[] _Objects = null;
		/// <summary>
		/// Gets an array of ScreenObjects that this screen is composed of.
		/// </summary>
        /// <remarks>This property is cached.</remarks>
		public ScreenObject[] Objects {
			get {
				// If objects for this screen haven't been extracted yet, extract them
				if(_Objects == null) {

					List<ScreenObject> objects = new List<ScreenObject>(8);
					int DataOffset = _Offset + 1;

					// While...
					//    We haven't hit the end-of-objects marker...
					//    and there is enough data left to represent an object {
					while(
						(DataOffset < _Data.Length && _Data[DataOffset] != RomValues.EndOfObjects && _Data[DataOffset] != RomValues.EndOfScreenData)
						&& (DataOffset + 2 < _Data.Length)) {

						// Load the object
						objects.Add(new ScreenObject(_Data, DataOffset));

						// Seek to next object
						DataOffset += 3;
					}

					_Objects = objects.ToArray();

				}
				return _Objects;
			}
		}
		/// <summary>
		/// Discards the cached list of ScreenObjects. 
		/// </summary>
		/// <remarks>This should be done when the number
		/// or memory location of the screen object data changes. This does not need 
		/// to be done when a screen object is merely modified because the 
		/// actual data is not cached, only a reference to the data is chached.
		///</remarks>
		public void InvalidateScreenObjects() {
			_Objects = null;
		}

		ScreenDoor[] _Doors = null;
		/// <summary>
		/// Gets an array of ScreenDoor objects that represents the doors of this room.
		/// </summary>
		public ScreenDoor[] Doors {
			get {
				// Create and cache the door references if it hasnt been done already
				if(_Doors == null) {
					LoadEnemyData();
				} // if

				return _Doors;
			} // get
		}

        /// <summary>
        /// Gets the index of a screen item.
        /// </summary>
        /// <param name="i">The item to find the index of.</param>
        /// <returns>The index of the specified item, or -1 if the item is not found.</returns>
        public int GetIndex(ScreenItem i) {
            if(i.IsNothing) return -1;
            if(i.SelectionType == ScreenItem.ObjectType.Enemy)
                return GetIndex(i.ThisEnemy);
            // Todo: Screen.GetIndex(ScreenDoor) and Screen.GetIndex(ScreenObject).
            return -1;
        }

        /// <summary>
        /// Gets the index of a screen item.
        /// </summary>
        /// <param name="e">The item to find the index of.</param>
        /// <returns>The index of the specified item, or -1 if the item is not found.</returns>
        public int GetIndex(ScreenEnemy e) {
            if(e.IsNothing) return -1;

            for(int i = 0; i < Enemies.Length; i++) {
                if(e.Equals(_Enemies[i])) 
                    return i; // Return index of match
            }

            return -1; // Not found
        }

        /// <summary>
        /// Loads all objects in the enemy data section. This includes doors, enemies, tourain access bridge, and possible other unknown data.
        /// </summary>
		private void LoadEnemyData() {
			List<ScreenDoor> doors = new List<ScreenDoor>(2);
            List<ScreenEnemy> enemies = new List<ScreenEnemy>(5);

			// Calculate offset (CAT + start of data + objects (3b ea.) + FD)
			int DataOffset = _Offset + Objects.Length * 3 + 1;
			// If there is no enemy data section, return empty arrays.
            if(_Data[DataOffset] == RomValues.EndOfScreenData) {
				_Doors = new ScreenDoor[] { };
				_Enemies = new ScreenEnemy[] { };
				return;
			}

			DataOffset++;

            while(DataOffset + 1 < _Data.Length && _Data[DataOffset] != RomValues.EndOfScreenData) {
                Byte dataMarker = _Data[DataOffset];
                if(dataMarker == RomValues.DoorMarker) {
                    doors.Add(new ScreenDoor(_Data, DataOffset));
                    DataOffset += 2;
                } else if((dataMarker & 0x0F) == RomValues.EnemyMarker || (dataMarker & 0x0F) == RomValues.RespawnEnemyMarker) {
                    enemies.Add(new ScreenEnemy(_Data, DataOffset));
                    DataOffset += 3;
                } else if(dataMarker == RomValues.TourainAccessBridgeMarker) {
                    hasBridge = true;
                    DataOffset++;
                } else {
                    break; // Invalid data, stop parsing enemy data
                }
            }


            //// While there is enough data to represent a door and there are door markers
            //while(DataOffset + 1 < _Data.Length && _Data[DataOffset] == RomValues.DoorMarker) {
            //    doors.Add(new ScreenDoor(_Data, DataOffset));
            //    DataOffset += 2;
            //} // while

            //// Cache doors

            //// If there are no enemies (we encounter an FF immediately after obj or door data)
            //// return an empty array
            //if(_Data[DataOffset] != RomValues.EndOfScreenData) {
            //    while(DataOffset + 2 < _Data.Length && _Data[DataOffset] != RomValues.EndOfScreenData) {
            //        enemies.Add(new ScreenEnemy(_Data, DataOffset));
            //        DataOffset += 3;
            //    }

            //}

            _Doors = doors.ToArray();
            _Enemies = enemies.ToArray();

		}

        private bool hasBridge;

        /// <summary>
        /// Gets whether this screen has an access bridge (a bridge appears if
        /// both bosses have been beaten).
        /// </summary>
        public bool HasBridge {
            get {
                // Load screen data if it is not loaded
                if(_Doors == null)
                    LoadEnemyData();

                return hasBridge; 
            }
        }


		/// <summary>
		/// Discards the cached list of ScreenDoors. 
		/// </summary>
		/// <remarks>This should be done when the number
		/// or memory location of the screen door data changes. This does not need 
		/// to be done when a screen door is merely modified because the 
		/// actual data is not cached, only a reference to the data is chached.
		///</remarks>
		public void InvalidateScreenDoors() {
			_Doors= null;
		}

		ScreenEnemy[] _Enemies = null;
		/// <summary>
		/// Gets an array of ScreenEnemy objects that represents the enemies in this room.
		/// </summary>
		public ScreenEnemy[] Enemies {
			get {
				if(_Enemies == null) {
					LoadEnemyData();
				}

				return _Enemies;
			}
		}
		/// <summary>
		/// Discards the cached list of ScreenEnemies. 
		/// </summary>
		/// <remarks>This should be done when the number
		/// or memory location of the screen enemy data changes. This does not need 
		/// to be done when a screen enemy is merely modified because the 
		/// actual data is not cached, only a reference to the data is chached.
		///</remarks>
		public void InvalidateScreenEnemies() {
			_Enemies = null;
		}
	} // Class
} // Namespace
