using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using Editroid.ROM;
using Editroid.Graphic;

namespace Editroid
{
	/// <summary>
	/// Provides an object-oriented and organized mechanism for accessing raw ROM data
	/// </summary>
	public class Rom
	{
		/// <summary>
		/// The raw data of the ROM
		/// </summary>
		public byte[] data;

        const int RomSize = 131088;
		/// <summary>
		/// Creates a Editroid.Rom from a stream
		/// </summary>
		/// <param name="s">A stream that contains a Metroid ROM image</param>
		public Rom(Stream s) {
			try {
				data = new byte[RomSize];
				s.Read(data, 0, data.Length);
			}
			catch(Exception Ex) {
				MessageBox.Show("Caught " + Ex.GetType().ToString() + " - " + Ex.Message);
				throw;
			}

			LoadLevels();
            //MakeGameItems();
            passwordData = new PasswordData(data);
		}

        private PasswordData passwordData;

        /// <summary>
        /// Gets/sets password data for this level.
        /// </summary>
        public PasswordData PasswordData {
            get { return passwordData; }
            set { passwordData = value; }
        }

		/// <summary>
		/// Saves the edited data of this ROM to a stream.
		/// </summary>
		/// <param name="s">The stream to output data to.</param>
		public void SaveTo(Stream s) {
			s.Write(data, 0, data.Length);
		}

		private void LoadLevels() {
			_Brinstar = new Level(this, Levels.Brinstar);
			_Ridley = new Level(this, Levels.Ridley);
			_Norfair = new Level(this, Levels.Norfair);
			_Kraid = new Level(this, Levels.Kraid);
			_Tourain = new Level(this, Levels.Tourain);
		}


		const int RoomBaseOffset = 0x254E;
		/// <summary>
		/// Gets which screen to use for a specified map position
		/// </summary>
		/// <param name="X">The X-coordinate on the map</param>
		/// <param name="Y">The Y-coordinate on the map</param>
		/// <returns>The room index of the specified room</returns>
		public byte GetRoomIndex(int X, int Y) {
			return GetRoomIndex(X + Y * 0x20);
		}
		/// <summary>
		/// Gets which screen to use for a specified map position
		/// </summary>
		/// <param name="roomOffset">The offset of the room</param>
		/// <returns>The room index of the specified room</returns>
		public byte GetRoomIndex(int roomOffset) {
			return data[RoomBaseOffset + roomOffset];
		}

		/// <summary>
		/// Sets which screen to use for a specified map position
		/// </summary>
		/// <param name="X">The X-coordinate of the room</param>
		/// <param name="Y">The Y-coordinate of the room</param>
		/// <param name="value">The index of the room</param>
		public void SetRoomIndex(int X, int Y, byte value) {
			SetRoomIndex(X + Y * 0x20, value);
		}
		/// <summary>
		/// Sets which screen to use for a specified map position
		/// </summary>
		/// <param name="Value">The index of the room</param>
		/// <param name="roomOffset">The index of the room</param>
		public void SetRoomIndex(int roomOffset, byte Value) {
			data[RoomBaseOffset + roomOffset] = Value;
		}

		Level _Brinstar, _Ridley, _Kraid, _Tourain, _Norfair;
		/// <summary>
		/// Brinstar level data.
		/// </summary>
		public Level Brinstar { get { return _Brinstar; } }
		/// <summary>
		/// Ridley level data.
		/// </summary>
		public Level Ridley { get { return _Ridley; } }
		/// <summary>
		/// Kraid level data.
		/// </summary>
		public Level Kraid { get { return _Kraid; } }
		/// <summary>
		/// Tourain level data.
		/// </summary>
		public Level Tourain { get { return _Tourain; } }
		/// <summary>
		/// Norfair level data.
		/// </summary>
		public Level Norfair { get { return _Norfair; } }

		/// <summary>
		/// Returns a Editroid.Level based on the specified value.
		/// </summary>
		/// <param name="level">The level to get</param>
		/// <returns>A Editroid.Level based on the specified value</returns>
		public Level GetLevel(Levels level) {
			switch(level) {
				case Levels.Brinstar:
					return _Brinstar;
				case Levels.Kraid:
					return _Kraid;
				case Levels.Norfair:
					return _Norfair;
				case Levels.Ridley:
					return _Ridley;
				case Levels.Tourain:
					return _Tourain;
                default:
                    return _Brinstar;
			}
		}

        /// <summary>
        /// Converts game text into a unicode string.
        /// </summary>
        /// <param name="offset">ROM offset to load data from.</param>
        /// <param name="length">The number of characters to load.</param>
        /// <returns>A string representation of ROM text.</returns>
        /// <remarks>Invalid or unknown data will be converted to underscores.</remarks>
        public string GetRomText(int offset, int length) {
            char[] result = new char[length];
            for(int i = 0; i < length; i++) {
                byte b = data[i + offset];
                if(b >= 0 && b <= 9) // Digit
                    result[i] = (char)((int)'0' + b);
                else if(b >= 0xA && b <= 0x23) // Letter
                    result[i] = (char)((int)'A' + b - 0xA);
                else if(b >= 0x24 && b <= 0x3D) // Lowercase letter
                    result[i] = (char)((int)'a' + b - 0x24);
                else if(b == 0x3E) // Question Mark
                    result[i] = '?';
                else if(b == 0x3F) // Dash
                    result[i] = '-';
                else if(b == 0xFF) // Space
                    result[i] = ' ';
                else if(b == 0x8F) // Copyright
                    result[i] = ''; 
                else // unknown
                    result[i] = '_';

            }

            return new string(result);
        }
        /// <summary>
        /// Converts a unicode string into Metroid formatted text and writes it to the ROM.
        /// </summary>
        /// <param name="offset">The offset to write the string to.</param>
        /// <param name="text">The text to write to the ROM.</param>
        /// <remarks>Underscores will be ignored and the ROM data in their place
        /// will remain. This is done so that invalid or unknown data will remain 
        /// intact in a round-trip operation, but can have unforseen consequences
        /// if data is inserted or removed from the a string during a round-trip
        /// operation, or if text happens to have underscores in it.</remarks>
        public void SetRomText(int offset, string text) {
            for(int i = 0; i < text.Length; i++) {

                char c = text[i];

                if(c >= '0' && c <= '9') // Digit
                    data[i + offset] = (byte)(c - '0');
                else if(c >= 'A' && c <= 'Z') // Letter
                    data[i + offset] = (byte)(c - 'A' + 0xA);
                else if(c >= 'a' && c <= 'z') // Lowercase letter
                    data[i + offset] = (byte)(c - 'a' + 0x24);
                else if(c == '?')
                    data[i + offset] = 0x3E;
                else if(c == '-')
                    data[i + offset] = 0x3F;
                else if(c == ' ')
                    data[i + offset] = 0xFF;
                else if(c == '')
                    data[i + offset] = 0x8F;
                else if(c == '_') // Unknown
                    ; // Don't modify value
                else // Unknown
                    data[i + offset] = 0x3E;
            }
        }
        /// <summary>ROM offset of the "PUSH START BUTTON" text.</summary>
        public const int TitleText_Line1_Offset = 0x514;
        /// <summary>ROM offset of copyright line in title screen.</summary>
        public const int TitleText_Line2_Offset = 0x52B;
        /// <summary>Length of the "PUSH START BUTTON" text.</summary>
        public const int TitleText_Line1_Length = 17;
        /// <summary>Length of the copyright line in the title screen.</summary>
        public const int TitleText_Line2_Length = 15;

        /// <summary>
        /// Gets or sets the text displayed below "Metroid" on the title screen.
        /// </summary>
        public String TitleText_Line1 {
            get {
                return GetRomText(TitleText_Line1_Offset, TitleText_Line1_Length);
            }
            set {
                if(value == null) value = string.Empty;
                if(value.Length > TitleText_Line1_Length)
                    value = value.Substring(0, TitleText_Line1_Length);

                SetRomText(TitleText_Line1_Offset, value.PadRight(TitleText_Line1_Length, ' '));
            }
        }
        /// <summary>
        /// Gets or sets the copyright line on the title screen.
        /// </summary>
        public String TitleText_Line2 {
            get {
                return GetRomText(TitleText_Line2_Offset, TitleText_Line2_Length);
            }
            set {
                if(value == null) value = string.Empty;
                if(value.Length > TitleText_Line2_Length)
                    value = value.Substring(0, TitleText_Line2_Length);

                SetRomText(TitleText_Line2_Offset, value.PadRight(TitleText_Line2_Length, ' '));
            }
        }

		/// <summary>
		/// Returns a Editroid.Level based on the specified value.
		/// </summary>
		/// <param name="level">The level to get</param>
        /// <param name="defaultLevel">The level to use if the level parameter is invalid.</param>
		/// <returns>A Editroid.Level based on the specified value</returns>
		public Level GetLevel(Levels level, Levels defaultLevel) {
            if(level == Levels.Invalid) level = defaultLevel;

			switch(level) {
				case Levels.Brinstar:
					return _Brinstar;
				case Levels.Kraid:
					return _Kraid;
				case Levels.Norfair:
					return _Norfair;
				case Levels.Ridley:
					return _Ridley;
				case Levels.Tourain:
					return _Tourain;
			}

			throw new ArgumentException("Value for \"level\" invalid.", "level");
		}

        
	}

    



}
