using System;
using System.Collections.Generic;
using System.Text;

namespace Editroid
{
    /// <summary>
    /// Encapsulates an item that may appear in a screen, i.e. an object, an
    /// enemy, or a door
    /// </summary>
    public class ScreenItem
    {
        ScreenObject thisObject;
        ScreenDoor thisDoor;
        ScreenEnemy thisEnemy;

        /// <summary>
        /// Creates an instance of this object that does not refer to an object.
        /// </summary>
        public ScreenItem() { }
        /// <summary>
        /// Creates a ScreenItem object.
        /// </summary>
        /// <param name="o">The object the ScreenItem will refer to.</param>
        public ScreenItem(ScreenObject o) { thisObject = o; }
        /// <summary>
        /// Creates a ScreenItem object.
        /// </summary>
        /// <param name="d">The object the ScreenItem will refer to.</param>
        public ScreenItem(ScreenDoor d) { thisDoor = d; }
        /// <summary>
        /// Creates a ScreenItem object.
        /// </summary>
        /// <param name="e">The object the ScreenItem will refer to.</param>
        public ScreenItem(ScreenEnemy e) { thisEnemy = e; }

        /// <summary>
        /// Gets the type of object this ScreenItem refers to.
        /// </summary>
        public ObjectType SelectionType {
            get {
                if(!(thisEnemy.IsNothing))
                    return ObjectType.Enemy;
                if(!(thisDoor.IsNothing))
                    return ObjectType.Door;
                if(!(thisObject.IsNothing))
                    return ObjectType.Object;

                return ObjectType.Nothing;
            }
        }

        /// <summary>
        /// Gets a string representation of this object.
        /// </summary>
        /// <returns>Gets a string representation of this object.</returns>
        public override string ToString() {
            string result = "(ScreenItem)";
            if(!thisEnemy.IsNothing)
                return result + thisEnemy.ToString();
            if(!thisDoor.IsNothing)
                return result + thisDoor.ToString();
            if(!thisObject.IsNothing)
                return result + thisObject.ToString();

            return "ScreenItem: [null]";
        }

        /// <summary>
        /// Gets whether the specified object is considered equal to this object.
        /// </summary>
        /// <param name="obj">The object to compare for equality.</param>
        /// <returns>A boolean value indicating equality.</returns>
        /// <remarks>A comparison to a null reference will always return a false value.</remarks>
        public override bool Equals(object obj) {
            ScreenItem s = obj as ScreenItem;
            if(s == null)
                return false;

            return Equals(s);
        }

        /// <summary>
        /// Gets whether the specified object is considered equal to this object.
        /// </summary>
        /// <param name="s">The object to compare for equality.</param>
        /// <returns>A boolean value indicating equality.</returns>
        /// <remarks>A comparison to a null reference will always return a false value.</remarks>
        public bool Equals(ScreenItem s) {
            if(s == null) return false;

            return s.thisDoor.Equals(thisDoor)
                || s.thisEnemy.Equals(thisEnemy)
                || s.thisObject.Equals(thisObject);
        }

        /// <summary>
        /// Gets or sets the door, enemy, or screen object this object represents.
        /// </summary>
        public object Item {
            get {
                switch(SelectionType) {
                    case ObjectType.Door:
                        if(thisDoor.IsNothing) return null;
                        return thisDoor;
                    case ObjectType.Enemy:
                        if(thisEnemy.IsNothing) return null;
                        return thisEnemy;
                    case ObjectType.Object:
                        if(thisObject.IsNothing) return null;
                        return thisObject;
                    default:
                        return null;
                }
            }
            set {
                if(value is ScreenEnemy) {
                    ThisEnemy = (ScreenEnemy)value;
                } else if(value is ScreenDoor) {
                    ThisDoor = (ScreenDoor)value;
                } else if(value is ScreenObject) {
                    ThisObject = (ScreenObject)value;
                } else if(value == null) {
                    ThisEnemy = ScreenEnemy.Nothing;
                } else {
                    throw new InvalidOperationException("Object is not a valid screen item.");
                }
            }
        }

        /// <summary>
        /// Gets a value which, when true, indicates that this object does not currently represent a value.
        /// This property can be set to true explicitly to clear its data, but it can not
        /// be set to false explicitly.
        /// </summary>
        public bool IsNothing {
            get {
                return thisObject.IsNothing &&
                    thisEnemy.IsNothing &&
                    thisDoor.IsNothing;
            }
            set {
                if(!value) throw new InvalidOperationException("Selection.IsNothing can not be set to false explicitly.");

                thisDoor = ScreenDoor.Nothing;
                thisEnemy = ScreenEnemy.Nothing;
                thisObject = ScreenObject.Nothing;
            }
        }

        /// <summary>
        /// Gets or sets the enemy that this selection represents.
        /// </summary>
        public ScreenObject ThisObject {
            get {
                return thisObject;
            }
            set {
                if(value.IsNothing) return;

                thisObject = value;
                thisDoor = ScreenDoor.Nothing;
                thisEnemy = ScreenEnemy.Nothing;
            }
        }

        /// <summary>
        /// Gets or sets the enemy that this selection represents.
        /// </summary>
        public ScreenEnemy ThisEnemy {
            get {
                return thisEnemy;
            }
            set {
                thisEnemy = value;
                thisDoor = ScreenDoor.Nothing;
                thisObject = ScreenObject.Nothing;
            }
        }

        /// <summary>
        /// Gets or sets the door that this selection represents.
        /// </summary>
        public ScreenDoor ThisDoor {
            get {
                return thisDoor;
            }
            set {
                thisDoor = value;
                thisEnemy = ScreenEnemy.Nothing;
                thisObject = ScreenObject.Nothing;
            }
        }

        /// <summary>
        /// Gets/sets the palette data if this object represents
        /// a screen item with palette data.
        /// </summary>
        public int Palette {
            get {
                if(!(thisEnemy.IsNothing))
                    return thisEnemy.SpritePal;
                else if(!(thisObject.IsNothing))
                    return thisObject.PalData;
                else
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
            }
            set {
                if(!(thisEnemy.IsNothing))
                    thisEnemy.SpritePal = value;
                else if(!(thisObject.IsNothing))
                    thisObject.PalData = (byte)value;
                else
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
            }
        }

        /// <summary>
        /// Gets/sets the X value of the currently selected enemy or object.
        /// </summary>
        public int X {
            get {
                if(!(thisEnemy.IsNothing)) {
                    return thisEnemy.X;
                } else if(!(thisObject.IsNothing)) {
                    return thisObject.X;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
            set {
                if(!(thisEnemy.IsNothing)) {
                    thisEnemy.X = value;
                } else if(!(thisObject.IsNothing)) {
                    thisObject.X = value;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
        }
        /// <summary>
        /// Gets/sets the Y value of the currently selected enemy or object.
        /// </summary>
        public int Y {
            get {
                if(!(thisEnemy.IsNothing)) {
                    return thisEnemy.Y;
                } else if(!(thisObject.IsNothing)) {
                    return thisObject.Y;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
            set {
                if(!(thisEnemy.IsNothing)) {
                    thisEnemy.Y = value;
                } else if(!(thisObject.IsNothing)) {
                    thisObject.Y = value;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
        }

        /// <summary>
        /// Gets/sets the index value which defines the type of object or enemy this selection represents.
        /// </summary>
        public int Index {
            get {
                if(!(thisEnemy.IsNothing)) {
                    return thisEnemy.EnemyType;
                } else if(!(thisObject.IsNothing)) {
                    return thisObject.Index;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
            set {
                if(!(thisEnemy.IsNothing)) {
                    thisEnemy.EnemyType = value;
                } else if(!(thisObject.IsNothing)) {
                    thisObject.Index = value;
                } else {
                    throw new InvalidOperationException("Can not get/set this value for the current selection.");
                }
            }
        }

        /// <summary>
        /// Represents a type of object in a game screen.
        /// </summary>
        public enum ObjectType
        {
            /// <summary>Represents a background object.</summary>
            Object,
            /// <summary>Represents an opening door.</summary>
            Door,
            /// <summary>Represents an enemy.</summary>
            Enemy,
            /// <summary>Represents nothing or an unspecified value.</summary>
            Nothing
        }
    }
}
