using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using Editroid.Graphic;
using Editroid.Properties;
using iLab.Sbl;
using Microsoft.Win32;
using System.Security;

namespace Editroid
{
    internal partial class frmMain:Form
    {
        Rom GameRom;
        RegistryKey settingsKey;

        const string _SaveMapToROM = "SaveMapToRom";
        const string _SaveBackupAsNES = "SaveBackupAsNES";
        const string _FeedbackUrl = "FeedbackUrl";

        MapControl MapEditor;
        ScreenControl ScreenEditor { get { return ScreenGrid.SelectedScreen; } }

        ScreenItem currentItem = new ScreenItem();
        Screen CurrentScreen { get { return ScreenEditor.CurrentScreen; } }
        Level currentLevel { get { return GameRom.GetLevel(ScreenEditor.LevelIndex); } }

        #region Dialog Windows
        frmHelp _helpForm;
        frmHelp HelpForm {
            get {
                if(_helpForm == null || _helpForm.IsDisposed)
                    _helpForm = new frmHelp();

                return _helpForm;
            }
        }

        IpsForm ipsForm;
        IpsForm IpsForm {
            get {
                if(ipsForm == null || ipsForm.IsDisposed)
                    ipsForm = new IpsForm();

                return ipsForm;
            }
        }
        #endregion


        public frmMain() {
            // General initialization
            InitializeComponent();
            MapEditor = ScreenGrid.MapEditor;
            LoadSettings();

            // Load default map
            MemoryStream m = new MemoryStream(Editroid.Properties.Resources.DefaultMap);
            ScreenGrid.MapEditor.TryLoadData(m, true);
            m.Close();

            mapMaker.RoomDrawn += new EventHandler(OnMapRoomDrawn);

            feedbackUrl = txtFeedback.Text;

            GlobalEventManager.Manager.StructureChanged += new EventHandler<LevelEventArgs>(Global_StructureChanged);
            GlobalEventManager.Manager.ComboChanged += new EventHandler<ComboEventArgs>(Global_ComboChanged);
        }

        void Global_StructureChanged(object sender, LevelEventArgs e) {
            UpdateScreensShowing(e.Level);
        }

        void Global_ComboChanged(object sender, LevelEventArgs e) {
            UpdateScreensShowing(e.Level);
        }

        /// <summary>Refreshes any screen editors displaying the specified level.</summary>
        /// <param name="level">The level to update.</param>
        private void UpdateScreensShowing(Levels level) {
            foreach(ScreenControl screen in ScreenGrid.Screens) {
                if(screen.LevelIndex == level)
                    screen.ShowScreen();
            }
        }

        protected override void OnClosing(CancelEventArgs e) {
            base.OnClosing(e);

            SaveSettings();
            GlobalEventManager.Manager.StructureChanged -= new EventHandler<LevelEventArgs>(Global_StructureChanged);
            GlobalEventManager.Manager.ComboChanged -= new EventHandler<ComboEventArgs>(Global_ComboChanged);
        }


        #region Settings
        private void LoadSettings() {
            try {
                settingsKey = Registry.CurrentUser.OpenSubKey("Software\\Ilab\\Editriod");
                if(settingsKey == null)
                    SaveMapToROM = true;
                else {
                    string mapSetting = settingsKey.GetValue(_SaveMapToROM) as string;
                    if(mapSetting == null || mapSetting.ToUpper() == "FALSE" || mapSetting.ToUpper() == "NO")
                        SaveMapToROM = false;
                    else
                        SaveMapToROM = true;
                
                    string backupSettings = settingsKey.GetValue(_SaveBackupAsNES) as string;
                    if(backupSettings == null || backupSettings.ToUpper() == "FALSE" || backupSettings.ToUpper() == "NO")
                        UseNesForBackup = false;
                    else
                        UseNesForBackup = true;

                    feedbackUrl = settingsKey.GetValue(_FeedbackUrl, "http://ilab.ahemm.org/editroid/feedback.php").ToString();
                    txtFeedback.Text = feedbackUrl;
                }
            }
            catch(SecurityException) {
                SaveMapToROM = true;
            }
        }
        private void SaveSettings() {
            try {
                settingsKey = Registry.CurrentUser.CreateSubKey("Software\\Ilab\\Editriod");
                if(settingsKey != null) {
                    settingsKey.SetValue(_SaveMapToROM, SaveMapToROM.ToString());
                    settingsKey.SetValue(_SaveBackupAsNES, UseNesForBackup.ToString());
                    settingsKey.SetValue(_FeedbackUrl, feedbackUrl);
                }
            }
            catch(SecurityException) {
                // Ignore
            }
        }


        bool SaveMapToROM {
            get { return mnuSaveMapToROM.Checked; }
            set {
                mnuSaveMapToROM.Checked = value;
                mnuSaveMapToFile.Checked = !value;
            } 
        }
        bool UseNesForBackup {
            get {
                return mnuUseNES.Checked;
            }
            set {
                mnuUseNES.Checked = value;
                mnuUseBAK.Checked = !value;
            }
        }
        #endregion

        
        /// <summary>
        /// Indicates that editing should be disabled, for example,
        /// when a blank map room (index 0xFF) is selected).
        /// </summary>
        bool NoEditing = false;

        /// <summary>
        /// Updates all controls and menus related to screen editing, as well as the screen display.
        /// </summary>
        private void UpdateSelectionStatus() {
            if(currentLevel == null) return;

            //currentScreen = currentLevel.GetScreen(ScreenGrid.SelectedScreen.ScreenIndex);
            //NoEditing = GameRom.GetRoomIndex(MapEditor.SelectionX, MapEditor.SelectionY) == 0xFF;

            UpdateDoorMenus();

            mnuRemove.Enabled = !NoEditing && !currentItem.IsNothing;
            int freeSpace = currentLevel.FreeRoomData;
            if(NoEditing) freeSpace = 0;

            mnuAddEnemy.Enabled = mnuAddObject.Enabled = 
                freeSpace > (CurrentScreen.HasEnemyData ? 2 : 3);

            if(ScreenEditor.SelectedItem.IsNothing) {
                btnBringFront.Enabled =
                    btnSendBack.Enabled =
                    btnPalette.Enabled =
                    btnStructBack.Enabled =
                    btnStructNext.Enabled =
                    btnRespawn.Enabled =
                    false;
                btnRespawn.Enabled = false;
            } else if(ScreenEditor.SelectedItem.SelectionType == ScreenItem.ObjectType.Object) {
                btnBringFront.Enabled =
                    btnSendBack.Enabled =
                    btnPalette.Enabled =
                    btnStructBack.Enabled =
                    btnStructNext.Enabled =
                    true;
                btnRespawn.Enabled =
                    btnRespawn.Enabled 
                    = false;
            } else if(ScreenEditor.SelectedItem.SelectionType == ScreenItem.ObjectType.Enemy) {
                btnBringFront.Enabled =
                    btnSendBack.Enabled =
                    false;
                btnPalette.Enabled =
                    btnRespawn.Enabled = 
                    btnStructBack.Enabled =
                    btnStructNext.Enabled =
                    btnRespawn.Enabled =
                    true;

            }

        }

        private void UpdateDoorMenus() {
            mnuLeft10.Checked = mnuLeft5.Checked = mnuLeftNormal.Checked =
                mnuRight10.Checked = mnuRight5.Checked = mnuRightNormal.Checked =
                false;

            mnuLeftNone.Checked = mnuRightNone.Checked = true;

            ScreenDoor[] Doors = currentLevel.GetScreen(ScreenGrid.SelectedScreen.ScreenIndex).Doors;
            foreach(ScreenDoor d in Doors) {
                if(d.Side == DoorSide.Left)
                    switch(d.Type) {
                        case DoorType.Normal:
                            mnuLeftNormal.Checked = true;
                            mnuLeftNone.Checked = false;
                            break;
                        case DoorType.Missile:
                            mnuLeft5.Checked = true;
                            mnuLeftNone.Checked = false;
                            break;
                        case DoorType.TenMissile:
                            mnuLeft10.Checked = true;
                            mnuLeftNone.Checked = false;
                            break;
                    } else if(d.Side == DoorSide.Right)
                    switch(d.Type) {
                        case DoorType.Normal:
                            mnuRightNormal.Checked = true;
                            mnuRightNone.Checked = false;
                            break;
                        case DoorType.Missile:
                            mnuRight5.Checked = true;
                            mnuRightNone.Checked = false;
                            break;
                        case DoorType.TenMissile:
                            mnuRight10.Checked = true;
                            mnuRightNone.Checked = false;
                            break;
                    }
            }

            // Enable menu items to add doors if there is enough memory
            // or enable items if there is alread a door
            int neededSpace = 2; // space needed for doors
            if(!CurrentScreen.HasEnemyData) neededSpace++;

            mnuLeftNone.Enabled = mnuRightNone.Enabled = !NoEditing;
            mnuLeftNormal.Enabled = mnuLeft10.Enabled = mnuLeft5.Enabled =
                !NoEditing && (mnuLeftNone.Checked == false || currentLevel.FreeRoomData >= neededSpace);
            mnuRightNormal.Enabled = mnuRight10.Enabled = mnuRight5.Enabled =
                !NoEditing && (mnuRightNone.Checked == false || currentLevel.FreeRoomData >= neededSpace);
        }

        #region UI Events
        private void btnPalette_Click(object sender, EventArgs e) {
            if(ModifierKeys == Keys.Shift) {
                CurrentScreen.ColorAttributeTable = (byte)((CurrentScreen.ColorAttributeTable + 1) % 4);
                ScreenGrid.SelectedScreen.ShowScreen();
            } else {
                ScreenEditor.NextPalette();
                RedrawScreen();
                GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);
            }
        }

        private void btnBringFront_Click(object sender, EventArgs e) {
            ScreenEditor.BringObjToFront();
            RedrawScreen();

        }

        private void btnSendBack_Click(object sender, EventArgs e) {
            ScreenEditor.SendObjToBack();
            RedrawScreen();
        }

        private void ScreenEditor_SelectedObjectChanged(object sender, EventArgs e) {
            currentItem = ScreenGrid.SelectedScreen.SelectedItem;
            UpdateSelectionStatus();
        }

        private void ScreenGrid_KeyDown(object sender, KeyEventArgs e) {
            OnKeyDown(e);
        }

        bool IsShifting { get { return (ModifierKeys & Keys.Shift) == Keys.Shift; } }
        bool IsControlling { get { return (ModifierKeys & Keys.Control) == Keys.Control; } }
        protected override void OnKeyDown(KeyEventArgs e) {
            base.OnKeyDown(e);

            if(GameRom == null) return;

            CheckNumpad(e);

            if(e.KeyCode == Keys.Left) {
                if(IsShifting || IsControlling) {
                    // Move on map
                    if(MapEditor.SelectionX != 0) MapEditor.SelectionX -= 1;
                } else {
                    // Move object
                    ScreenEditor.MoveObject(-1, 0);
                    RedrawScreen();
                }
            } else if(e.KeyCode == Keys.Up) {
                if(IsShifting || IsControlling) {
                    // Move on map
                    if(MapEditor.SelectionY != 0) MapEditor.SelectionY -= 1;
                } else {
                    // Move object
                    ScreenEditor.MoveObject(0, -1);
                    RedrawScreen();
                }
            } else if(e.KeyCode == Keys.Right) {
                if(IsShifting || IsControlling) {
                    // Move on map
                    if(MapEditor.SelectionX != 31) MapEditor.SelectionX += 1;
                } else {
                    // Move object
                    ScreenEditor.MoveObject(1, 0);
                    RedrawScreen();
                }
            } else if(e.KeyCode == Keys.Down) {
                if(IsShifting || IsControlling) {
                    // Move on map
                    if(MapEditor.SelectionY != 31) MapEditor.SelectionY += 1;
                } else {
                    // Move object
                    ScreenEditor.MoveObject(0, 1);
                    RedrawScreen();
                }
            } else if(e.KeyCode == Keys.PageUp) {
                PrevScreen();
            } else if(e.KeyCode == Keys.PageDown) {
                NextScreen();
            } else if(e.KeyCode == Keys.OemMinus || e.KeyCode == Keys.Subtract) {
                ScreenEditor.SeekSelectedObjectType(-1);
                RedrawScreen();
                GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);
            } else if(e.KeyCode == Keys.Oemplus || e.KeyCode == Keys.Add) {
                ScreenEditor.SeekSelectedObjectType(1);
                RedrawScreen();
                GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);
            } else if(e.KeyCode == Keys.Home) {
                ScreenEditor.BringObjToFront();
                RedrawScreen();
            } else if(e.KeyCode == Keys.End) {
                ScreenEditor.SendObjToBack();
                RedrawScreen();
            } else if(e.KeyCode == Keys.P || e.KeyCode == Keys.Space) {
                ScreenEditor.NextPalette();
                RedrawScreen();
                GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);
            } else if((ModifierKeys & Keys.Shift) == Keys.Shift &&
                e.KeyCode == Keys.Oemtilde) {
                SetPalette(0);
            } else if((ModifierKeys & Keys.Shift) == Keys.Shift &&
                e.KeyCode == Keys.D1) {
                SetPalette(1);
            } else if((ModifierKeys & Keys.Shift) == Keys.Shift &&
                e.KeyCode == Keys.D2) {
                SetPalette(2);
            } else if((ModifierKeys & Keys.Shift) == Keys.Shift &&
                e.KeyCode == Keys.D3) {
                SetPalette(3);
            } else if(e.KeyCode == Keys.Tab) {
                if(IsControlling) {
                    ScreenEditor.ToggleSelectionType();
                } else if(IsShifting) {
                    ScreenEditor.SelectPrevious();
                } else {
                    ScreenEditor.SelectNext();
                }
                RedrawScreen();
            } else if(e.KeyCode == Keys.Back || e.KeyCode == Keys.Delete) {
                if(mnuRemove.Enabled)
                    RemoveSelection();
            } else if(e.KeyCode == Keys.Oemcomma) {
                if(mnuAddObject.Enabled)
                    mnuAddObject_Click(this, e);
            } else if(e.KeyCode == Keys.OemPeriod) {
                if(mnuAddEnemy.Enabled)
                    mnuAddEnemy_Click(this, e);
            } else if(getIntValue(e.KeyCode) != -1) {
                OnNumericInput(getIntValue(e.KeyCode));
                return;
            }

            numericInput.Clear();
        }

        #endregion

        bool showingDoorsOnMap;
        void ShowDoorsOnMap() {
            showingDoorsOnMap = true;
            for(int x = 0; x < 31; x++) {
                for(int y = 0; y < 31; y++) {
                    Levels roomlevel = ScreenGrid.MapEditor.GetLevel(x, y);
                    if(roomlevel != Levels.Invalid) {
                        Level levelData = GameRom.GetLevel(roomlevel);
                        Screen screenData = levelData.GetScreen(GameRom.GetRoomIndex(x, y));
                        ScreenDoor[] DoorData = screenData.Doors;

                        MapRoom doordata = (MapRoom)0;
                        for(int i = 0; i < DoorData.Length; i++) {
                            if(DoorData[i].Side == DoorSide.Left)
                                doordata |= MapRoom.LeftDoor;
                            else
                                doordata |= MapRoom.RightDoor;
                        }

                        ScreenGrid.MapEditor.SetDoor(x, y, doordata);
                    }
                }
            }
            ScreenGrid.MapEditor.Refresh();
        }
        void ClearDoorsOnMap() {
            if(showingDoorsOnMap) {
                showingDoorsOnMap = false;

                for(int x = 0; x < 31; x++) {
                    for(int y = 0; y < 31; y++) {
                        Levels roomlevel = ScreenGrid.MapEditor.GetLevel(x, y);
                        if(roomlevel != Levels.Invalid) {


                            ScreenGrid.MapEditor.SetDoor(x, y, MapRoom.Brinstar);
                        }
                    }
                }
                ScreenGrid.MapEditor.Refresh();
            }
        }


        /// <summary>
        /// Handles keyboard events for numpad (map editing).
        /// </summary>
        private void CheckNumpad(KeyEventArgs e) {
            switch(e.KeyCode) {
                case Keys.NumPad0:
                    SetCurrentScreenLevel(Levels.Invalid);
                    break;
                case Keys.NumPad1:
                    SetCurrentScreenLevel(Levels.Brinstar);
                    break;
                case Keys.NumPad2:
                    SetCurrentScreenLevel(Levels.Norfair);
                    break;
                case Keys.NumPad3:
                    SetCurrentScreenLevel(Levels.Ridley);
                    break;
                case Keys.NumPad4:
                    SetCurrentScreenLevel(Levels.Kraid);
                    break;
                case Keys.NumPad5:
                    SetCurrentScreenLevel(Levels.Tourain);
                    break;
                default:
                    return;
            }
        }

        #region Keyboard input for struct/enemy index
        List<byte> numericInput = new List<byte>();
        private void OnNumericInput(int p) {
            // Register the keypress
            numericInput.Add((byte)p);

            if(!currentItem.ThisObject.IsNothing) {
                // Remove least-recently typed digits until the value is within range
                while(GetNumericInputValue() > currentLevel.StructCount - 1) {
                    numericInput.RemoveAt(0);
                }

                // Update object
                ScreenObject o = currentItem.ThisObject;
                o.Index = GetNumericInputValue();

                // Refresh display
                RedrawScreen();
                UpdateSelectionStatus();
            }
            
        }

        /// <summary>Gets the value that has been typed in by the user.</summary>
        int GetNumericInputValue() {
            int result = 0;
            int multiple = 1;
            for(int i = numericInput.Count; i > 0; i--) {
                result += numericInput[i - 1] * multiple;
                multiple *= 0x10;
            }

            return result;
        }

        /// <summary>
        /// Gets the value of the specified key, or -1 if the key does not represent a value.
        /// </summary>
        int getIntValue(Keys key) {
            switch(key) {
                case Keys.D0:
                    return 0;
                case Keys.D1:
                    return 1;
                case Keys.D2:
                    return 2;
                case Keys.D3:
                    return 3;
                case Keys.D4:
                    return 4;
                case Keys.D5:
                    return 5;
                case Keys.D6:
                    return 6;
                case Keys.D7:
                    return 7;
                case Keys.D8:
                    return 8;
                case Keys.D9:
                    return 9;
                case Keys.A:
                    return 0xA;
                case Keys.B:
                    return 0xB;
                case Keys.C:
                    return 0xC;
                case Keys.D:
                    return 0xD;
                case Keys.E:
                    return 0xE;
                case Keys.F:
                    return 0xF;
            }

            return -1;
        }

        #endregion

        /// <summary>Sets the active screen editor to the specified level and ensures that the layout index is within range. The 
        /// display will be refreshed.</summary>
        /// <param name="screenLevel"></param>
        private void SetCurrentScreenLevel(Levels screenLevel) {
            MapRoom mapLevel;
            switch(screenLevel) {
                case Levels.Brinstar:
                    mapLevel = MapRoom.Brinstar;
                    break;
                case Levels.Kraid:
                    mapLevel = MapRoom.Kraid;
                    break;
                case Levels.Norfair:
                    mapLevel = MapRoom.Norfair;
                    break;
                case Levels.Ridley:
                    mapLevel = MapRoom.Ridley;
                    break;
                case Levels.Tourain:
                    mapLevel = MapRoom.Tourain;
                    break;
                default:
                    mapLevel = MapRoom.Blank;
                    break;
            }

            int mapX = ScreenEditor.MapLocation.X;
            int mapY = ScreenEditor.MapLocation.Y;

            // Update screen editor
            ScreenEditor.LevelIndex = screenLevel;

            if(mapLevel == MapRoom.Blank) {
                ScreenEditor.ScreenIndex = 0xFF;

                GameRom.SetRoomIndex(MapEditor.SelectionX, MapEditor.SelectionY, (byte)0xFF);

                // Update the screen editor more
                ScreenEditor.ScreenIndex = 0xFF;
            }else{
                int levelScreenCount = GameRom.GetLevel(screenLevel).ScreenCount;
                int screenIndex = GameRom.GetRoomIndex(MapEditor.SelectionX, MapEditor.SelectionY);
                // Verify that screen is within valid range
                if(screenIndex == 0xFF) {
                    screenIndex = 0;
                    GameRom.SetRoomIndex(MapEditor.SelectionX, MapEditor.SelectionY, (byte)screenIndex);
                    ScreenEditor.ScreenIndex = screenIndex;
                }
                if(screenIndex > levelScreenCount) {
                    screenIndex = levelScreenCount - 1;
                    GameRom.SetRoomIndex(MapEditor.SelectionX, MapEditor.SelectionY, (byte)screenIndex);
                    ScreenEditor.ScreenIndex = screenIndex;
                }

                // Update the screen editor more
                ScreenEditor.ScreenIndex = screenIndex;
            }

            // Update UI
            RedrawScreen();
            UpdateSelectionStatus();

            // Update the map editor
            MapEditor.SetImage(mapX, mapY, mapLevel);
        }

        private void SetPalette(int pal) {
            if(!currentItem.IsNothing) {
                ScreenEditor.SetPalette((byte)pal);
                RedrawScreen();
            }
        }

        private void btnAltPal_Click(object sender, EventArgs e) {
            ScreenGrid.UseAlternatePalette = btnAltPal.Checked;
            ScreenGrid.RedrawAllScreens();
            if(!(structureEditor == null || structureEditor.IsDisposed))
                structureEditor.UseAlternatePalette = btnAltPal.Checked;
        }

        private void btnStructBack_Click(object sender, EventArgs e) {
            ScreenEditor.SeekSelectedObjectType(-1);
            RedrawScreen();
            GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);

        }


        private void btnStructNext_Click(object sender, EventArgs e) {
            ScreenEditor.SeekSelectedObjectType(1);
            RedrawScreen();
            GlobalEventManager.Manager.OnObjectEdited(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);

        }

        /// <summary>Redraws the screen being edited and any other screen editors showing the same screen.</summary>
        void RedrawScreen() {
            ScreenGrid.RedrawAllInstances(ScreenEditor.LevelIndex, ScreenEditor.ScreenIndex);
        }

        const string mapExtension = ".MP";
        string GetMapFileName() {
            return Path.ChangeExtension(cdgOpenRom.FileName, mapExtension);
        }
        private void btnOpen_Click(object sender, EventArgs e) {
            if(cdgOpenRom.ShowDialog() == DialogResult.OK) {
                try {
                    using(FileStream gameStream = new FileStream(cdgOpenRom.FileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite)) {
                        OpenRom(gameStream, true);
                    }
                }
                catch(Exception ex) {
                    MessageBox.Show("Could not load the specified file (" + ex.GetType().ToString() + "- " + ex.Message + ")");
                    return;
                }
                EnableControls();
                MapEditor.Filter(GameRom);
            }
        }

        private void OpenRom(Stream gameStream, bool LoadMapFromFile) {
            GameRom = new Rom(gameStream);
            
            // Load map (preference in order of: embedded (in ROM), external (MP file), default)
            if(MapEditor.TryLoadData(gameStream) == false) {
                if(File.Exists(GetMapFileName()) && LoadMapFromFile) {
                    MapEditor.LoadData(GetMapFileName());
                } else {
                    byte[] defaultMap = Resources.DefaultMap;
                    MemoryStream defaultMapStream = new MemoryStream(defaultMap);
                    MapEditor.TryLoadData(defaultMapStream);
                }
            }

            //ScreenGrid.MapEditor.TryLoadData(gameStream);
            ScreenGrid.Rom = GameRom;
            SetGridSize(2, 1, 1, 0);
            MapEditor.SelectionSize = new Size(1, 1);
        }

        private void SetGridSize(int width, int height, int mapx, int mapy) {
            ScreenGrid.SetSize(width, height, mapx, mapy);
            ScreenGrid.Display(MapEditor.SelectionX, MapEditor.SelectionY);
            this.ClientSize = new Size(2 + 258 * width, 2 + 258 * height + toolz.Height);
        }

        private void EnableControls() {
            btnAltPal.Enabled = true;
            btnBringFront.Enabled = true;
            btnPalette.Enabled = true;
            btnSave.Enabled = true;
            btnSendBack.Enabled = true;
            btnStructBack.Enabled = true;
            btnStructNext.Enabled = true;
            btnNextScreen.Enabled = true;
            btnShowDoors.Enabled = true;
            btnPrevScreen.Enabled = true;
            ScreenEditor.Enabled = true;
            MapEditor.Enabled = true;
            MemoryMenu.Enabled = true;
            mnuStart.Enabled = true; // Forgot this line on the 1.2 release.

            foreach(object o in mnuEditors.DropDownItems) {
                ToolStripMenuItem i = o as ToolStripMenuItem;
                if(i != null && (i.Tag == null || !i.Tag.ToString().Contains("NoEnable")))
                    i.Enabled = true;
            }

            ScreenEditor.Visible = true;
            MapEditor.Visible = true;
            picTitle.Visible = false;
        }

        private void btnSave_Click(object sender, EventArgs e) {
            FileAttributes atts = File.GetAttributes(cdgOpenRom.FileName);
            if((atts & FileAttributes.ReadOnly) != 0) {
                if(MessageBox.Show("The ROM file is read-only. It can not be written to unless the read-only attribute is disabled. Would you like to do this now?", "Error Saving File", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes) {
                    File.SetAttributes(cdgOpenRom.FileName, atts & ~FileAttributes.ReadOnly);
                }
            }
            using(FileStream gameStream = new FileStream(cdgOpenRom.FileName, FileMode.Create, FileAccess.Write, FileShare.ReadWrite)) {

                GameRom.SaveTo(gameStream);
                SaveMap(gameStream, GetMapFileName());
            }
        }

        /// <summary>
        /// Saves the editor map for a ROM.
        /// </summary>
        /// <param name="gameStream">The filestream the game is saved to.</param>
        /// <param name="filename">The filename to use if the map is to be saved to an external file.</param>
        /// <remarks>If an extraneous map is found (there is one in both the ROM image and an external file) it will be deleted.</remarks>
        private void SaveMap(FileStream gameStream, string filename) {
            SaveMap(gameStream, filename, SaveMapToROM);
        }
        /// <summary>
        /// Saves the editor map for a ROM.
        /// </summary>
        /// <param name="gameStream">The filestream the game is saved to.</param>
        /// <param name="filename">The filename to use if the map is to be saved to an external file.</param>
        /// <param name="saveMapToRom">Whether to save to a stream or the ROM.</param>
        /// <remarks>If an extraneous map is found (there is one in both the ROM image and an external file) it will be deleted.</remarks>
        private void SaveMap(FileStream gameStream, string filename, bool saveMapToRom) {
            if(saveMapToRom) {
                if(File.Exists(filename) && MessageBox.Show("Delete external .MAP file?", "Editroid", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes) {
                    // Delete separate map file
                    try {
                        File.Delete(filename);
                    }
                    catch(Exception) {
                        // Ignore
                    }
                }

                SaveMapToStream(gameStream);
            } else {
                FileStream mapStream = null;
                try {
                    mapStream = new FileStream(filename,  FileMode.Create);
                    SaveMapToStream(mapStream);
                }
                catch(Exception ex) {
                    MessageBox.Show("The editor map could not be saved.", ex.GetType().ToString());
                }
                finally {
                    if(mapStream != null)
                        mapStream.Close();
                }
            }
        }

        private void SaveMapToStream(FileStream stream) {
            MapEditor.SaveData(stream);
        }



        private void btnPrevScreen_Click(object sender, EventArgs e) {
            PrevScreen();
        }
        private void btnNextScreen_Click(object sender, EventArgs e) {
            NextScreen();
        }


        private void PrevScreen() {
            SeekScreenIndex(-1);
        }

        /// <summary>Advances the screen displayed for the map slot being edited.</summary>
        private void NextScreen() {
            SeekScreenIndex(1);
        }

        private void SeekScreenIndex(int count) {
            // Determine next screen
            int newScreen = GameRom.GetRoomIndex(ScreenEditor.MapLocation.X, ScreenEditor.MapLocation.Y);
            newScreen += count;

            // Loop around if we get to the last screen
            if(newScreen >= GameRom.GetLevel(MapEditor.GetLevel()).ScreenCount)
                newScreen = 0;
            if(newScreen < 0)
                newScreen = GameRom.GetLevel(MapEditor.GetLevel()).ScreenCount - 1;

            // Update game map
            GameRom.SetRoomIndex(ScreenEditor.MapLocation.X, ScreenEditor.MapLocation.Y, (byte)newScreen);

            // Update editor
            ScreenEditor.ScreenIndex = newScreen;
            ScreenEditor.MakeDefaultSelection();
            RedrawScreen();

            //currentScreen = ScreenEditor.CurrentScreen;
            UpdateSelectionStatus();
        }

        private void btnShowDoors_Click(object sender, EventArgs e) {
            ShowDoorsOnMap();
        }

        private void mnuEditPalette_Click(object sender, EventArgs e) {
            frmPalette paletteEditor = new frmPalette();
            paletteEditor.Data = GameRom.data;
            paletteEditor.StartPosition = FormStartPosition.CenterScreen;
            paletteEditor.ShowDialog();
            ScreenGrid.RedrawAllScreens();

        }

        frmPointer pointerForm;
        private void mnuPointer_Click(object sender, EventArgs e) {
            if(pointerForm == null || pointerForm.IsDisposed)
                pointerForm = new frmPointer();

            pointerForm.Show();
        }

        frmComboEditor comboForm;
        private void mnuEditCombos_Click(object sender, EventArgs e) {
            if(comboForm == null || comboForm.IsDisposed)
                comboForm = new frmComboEditor(GameRom);

            comboForm.ShowDialog();
            comboForm.Close();
            ScreenGrid.RedrawAllScreens();

        }

        frmStruct structureEditor;

        private void mnuEditStructs_Click(object sender, EventArgs e) {
            if(structureEditor == null || structureEditor.IsDisposed) {
                structureEditor = new frmStruct();
                structureEditor.Owner = this;
                structureEditor.Rom = GameRom;
                structureEditor.UseAlternatePalette = btnPalette.Checked;
                structureEditor.LoadLevel(ScreenEditor.LevelIndex);
                structureEditor.Show();
                GlobalEventManager.Manager.OnObjectSelected(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);

            } else {
                structureEditor.Show();
                structureEditor.BringToFront();
                return;
            }


            //ScreenGrid.RedrawAllScreens();
        }

        private void mnuSaveMapTo_Clicked(object sender, EventArgs e) {
            SaveMapToROM = sender == mnuSaveMapToROM;
        }

        private void mnuPhysics_Click(object sender, EventArgs e) {
            ScreenEditor.ShowScreen(true);
        }

        private void MapEditor_RoomCleared(object sender, int X, int Y) {
            ClearDoorsOnMap();

            GameRom.SetRoomIndex(X, Y, 255);
            //MapEditor_SelectionChanged(sender, X, Y);
        }

        private void mnuBackup_Click(object sender, EventArgs e) {
            DateTime time = DateTime.Now;
            string Filename = time.Year.ToString() + "_" + time.Month.ToString().PadLeft(2, '0') + "_" + time.Day.ToString().PadLeft(2, '0') + "_" + time.Hour.ToString().PadLeft(2, '0') + time.Minute.ToString().PadLeft(2, '0') + time.Second.ToString().PadLeft(2, '0') +
                (UseNesForBackup ? ".nes" : ".bak");
            Filename = Path.Combine(
                Path.GetDirectoryName(cdgOpenRom.FileName),
                Filename);

            FileStream gameStream = new FileStream(Filename, FileMode.Create, FileAccess.Write);
            GameRom.SaveTo(gameStream);
            SaveMap(gameStream, null, true);
            gameStream.Close();

        }

        private void mnuUseNES_Click(object sender, EventArgs e) {
            UseNesForBackup = true;
        }

        private void mnuUseBAK_Click(object sender, EventArgs e) {
            UseNesForBackup = false;
        }

        private void btnAbout_Click(object sender, EventArgs e) {
            HelpForm.ShowAbout();
        }

        private void btnGettingStarted_Click(object sender, EventArgs e) {
            HelpForm.ShowGettingStarted();
        }

        private void btnHelp_Click(object sender, EventArgs e) {
            HelpForm.ShowHelp();
        }


        private frmPointers pointerExplorerForm;
        public frmPointers PointerExplorerForm {
            get {
                if(pointerExplorerForm == null || pointerExplorerForm.IsDisposed) {
                    pointerExplorerForm = new frmPointers();
                    pointerExplorerForm.SetRom(GameRom);
                }
                return pointerExplorerForm; 
            }
        }

        private void mnuPointerExplorer_Click(object sender, EventArgs e) {
            PointerExplorerForm.Show();
        }

        private void btnRespawn_Click(object sender, EventArgs e) {
            ScreenEditor.ToggleRespawn();
            RedrawScreen();
        }


        frmTitleText titleForm;
        private void mnuTitleText_Click(object sender, EventArgs e) {
            if(titleForm == null || titleForm.IsDisposed)
                titleForm = new frmTitleText();

            titleForm.Rom = GameRom;
            titleForm.ShowDialog();
            titleForm.Dispose();
        }

        MapMaker mapMaker = new MapMaker();
        Bitmap map = null;

        frmMapArea mapForm;
        frmMapStatus progress;
        private void mnuMap_Click(object sender, EventArgs e) {
            if(mapForm == null || mapForm.IsDisposed)
                mapForm = new frmMapArea();

            ShowDoorsOnMap();
            
            mapForm.MapControl = MapEditor;


            if(mapForm.ShowDialog() == DialogResult.OK) {
                // Load data needed for map
                mapMaker.Rom = GameRom;
                mapMaker.Scale = mapForm.DrawScale;
                mapMaker.MapArea = mapForm.MapSelection;
                mapMaker.LevelMap = MapEditor;
                mapMaker.ShowPhysics = mapForm.ShowPhysics;
                mapMaker.HighQuality = mapForm.Quality;
                mapMaker.HideEnemies = mapForm.HideEnemies;
                mapMaker.LevelFilter = mapForm.LevelFilter;
                mapMaker.FillEmptySpots = mapForm.FillEmptySpots;

                // Display progress form
                progress = new frmMapStatus();
                progress.RoomCount = mapMaker.MapArea.Width * mapMaker.MapArea.Height;
                progress.Show();
                Application.DoEvents();

                // Create Map
                MapThread.RunWorkerAsync();
            }

            ClearDoorsOnMap();
        }

        /// <summary>
        /// Called each time a map room is drawn when a map is being created to update progress.
        /// </summary>
        void OnMapRoomDrawn(object sender, EventArgs e) {
            MapThread.ReportProgress(0);
        }

        private void MapThread_DoWork(object sender, DoWorkEventArgs e) {
            map = mapMaker.Render();

        }

        private void MapThread_ProgressChanged(object sender, ProgressChangedEventArgs e) {
            progress.CompleteRoom();
        }

        private void MapThread_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
            progress.Close();
            
            // Display Map
            frmMap mapDisplay = new frmMap();
            mapDisplay.MapImage = map;
            mapDisplay.Show();
        }

        private void btnRemove_Click(object sender, EventArgs e) {
            //RemoveSelectedObject();
            ScreenEditor.RemoveSelectedItem();
        }


        // Todo: Move to ScreenEditor
        private void AddNewObject() {
            ScreenEditor.AddObject();

            ScreenObject[] objects = currentLevel.GetScreen(ScreenEditor.ScreenIndex).Objects;
            ScreenEditor.SelectedItem.ThisObject = objects[objects.Length - 1];
            GlobalEventManager.Manager.OnObjectSelected(this, ScreenEditor.LevelIndex, ScreenEditor.SelectedItem);

            ScreenGrid.ReloadLevel(ScreenEditor.LevelIndex);
        }

        private void AddNewEnemy() {
            ScreenEditor.AddEnemy();
            ScreenEditor.RefreshData();
            ScreenEditor.SelectedItem.ThisEnemy = currentLevel.GetScreen(ScreenEditor.ScreenIndex).Enemies[0];
            ScreenGrid.ReloadLevel(ScreenEditor.LevelIndex);
        }



        private void mnuRemove_Click(object sender, EventArgs e) {
            RemoveSelection();
        }

        private void RemoveSelection() {
            ScreenEditor.RemoveSelectedItem();
            ScreenGrid.ReloadLevel(ScreenEditor.LevelIndex);
        }

        private void mnuAddObject_Click(object sender, EventArgs e) {
            AddNewObject();

            UpdateSelectionStatus();
        }

        private void mnuAddEnemy_Click(object sender, EventArgs e) {
            AddNewEnemy();

            UpdateSelectionStatus();
        }

        private void LeftDoor_Changed(object sender, EventArgs e) {
            string TypeName  = (sender as ToolStripMenuItem).Text;

            switch(TypeName) {
                case "None":
                    ScreenEditor.DeleteLeftDoor();
                    break;
                case "Normal":
                    ScreenEditor.SetLeftDoor(DoorType.Normal);
                    break;
                case "5 Missile":
                    ScreenEditor.SetLeftDoor(DoorType.Missile);
                    break;
                case "10 Missile":
                    ScreenEditor.SetLeftDoor(DoorType.TenMissile);
                    break;
            }

            ScreenGrid.ReloadLevel(ScreenEditor.LevelIndex);
            UpdateSelectionStatus();
        }

        private void RightDoorChanged(object sender, EventArgs e) {
            string TypeName = (sender as ToolStripMenuItem).Text;

            switch(TypeName) {
                case "None":
                    ScreenEditor.DeleteRightDoor();
                    break;
                case "Normal":
                    ScreenEditor.SetRightDoor(DoorType.Normal);
                    break;
                case "5 Missile":
                    ScreenEditor.SetRightDoor(DoorType.Missile);
                    break;
                case "10 Missile":
                    ScreenEditor.SetRightDoor(DoorType.TenMissile);
                    break;
            }

            ScreenGrid.ReloadLevel(ScreenEditor.LevelIndex);
            UpdateSelectionStatus();
        }

        private void mnuItemEditor_Click(object sender, EventArgs e) {
            ItemEditor itemEditor = new ItemEditor();
            itemEditor.MapImage = MapEditor.MapImage;
            itemEditor.MapSource = MapEditor;
            itemEditor.Rom = GameRom;
            itemEditor.ShowDialog();
            itemEditor.Dispose();
        }

        frmPassword passEditor;
        private void mnuPasswordEditor_Click(object sender, EventArgs e) {
            if(passEditor == null || passEditor.IsDisposed) {
                passEditor = new frmPassword();
                passEditor.Rom = GameRom;
                passEditor.MapEditor = MapEditor;
            }

            passEditor.ShowDialog();
            passEditor.Dispose();
        }

        private void mnuStart_Click(object sender, EventArgs e) {
            currentLevel.StartScreenX = MapEditor.SelectionX;
            currentLevel.StartScreenY = MapEditor.SelectionY;
        }

        private void MapEditor_selectedRoomModified(object sender, EventArgs e) {
            Levels level = MapEditor.GetLevel(MapEditor.SelectionX, MapEditor.SelectionY);
            if(level == Levels.Invalid) level = Levels.Brinstar;
            ScreenEditor.LevelIndex = level;

            RedrawScreen();
        }

        private void MapEditor_RoomSet(object sender, int X, int Y) {
            if(MapEditor.GetLevel(X, Y) != Levels.Invalid && GameRom.GetRoomIndex(X, Y) == 0xFF) {
                GameRom.SetRoomIndex(X, Y, 0);
            }
        }

        static string feedbackUrl;
        public static string FeedbackUrl { get { return feedbackUrl; } }

        private void txtFeedback_TextChanged(object sender, EventArgs e) {
            feedbackUrl = txtFeedback.Text;    
        }

        private void mnuFeedback_Click(object sender, EventArgs e) {
            HelpForm.ShowFeedback();
        }

        private void mnuIPS_Click(object sender, EventArgs e) {
            IpsForm.CurrentRom = GameRom;
            IpsForm.ShowDialog();
            if(IpsForm.CurrentPatched) {
                byte[] data = GameRom.data;
                MemoryStream s = new MemoryStream(data);
                OpenRom(s, false); // Re-loads the current ROM to reflect changes. Reloaded rom will still have same filename
                s.Close();

                MapEditor.Filter(GameRom);
            }
        }

        private void mnuDefaultGrid_Click(object sender, EventArgs e) {
            SetGridSize(2, 1, 1, 0);
            mnuDefaultGrid.Checked = true;
            mnuExpandedGrid.Checked = false;
            mnuHugeGrid.Checked = false;
            MapEditor.SelectionSize = new Size(1, 1);
        }

        private void mnuExpandedGrid_Click(object sender, EventArgs e) {
            SetGridSize(2, 2, 1, 1);
            mnuDefaultGrid.Checked = false;
            mnuExpandedGrid.Checked = true;
            mnuHugeGrid.Checked = false;
        }

        private void mnuHugeGrid_Click(object sender, EventArgs e) {
            SetGridSize(3, 3, 2, 2);
            mnuDefaultGrid.Checked = false;
            mnuExpandedGrid.Checked = false;
            mnuHugeGrid.Checked = true;
        }

        private void ScreenGrid_SelectedObjectDragged(object sender, EventArgs e) {
            RedrawScreen();
        }


    }
}